"""
HGNet v1, adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/hgnet.py

Reference document:
https://github.com/PaddlePaddle/PaddleClas/blob/release/2.5.1/docs/en/models/PP-HGNet_en.md
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import StochasticDepth

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class LearnableAffineBlock(nn.Module):
    def __init__(self, scale_value: float, bias_value: float) -> None:
        super().__init__()
        self.scale = nn.Parameter(torch.tensor([scale_value]), requires_grad=True)
        self.bias = nn.Parameter(torch.tensor([bias_value]), requires_grad=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.scale * x + self.bias


class Conv2dBNActivation(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        padding: tuple[int, int],
        groups: int,
        use_act: bool,
        use_lab: bool,
    ) -> None:
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            groups=groups,
            bias=False,
        )
        self.bn = nn.BatchNorm2d(out_channels)
        if use_act is True:
            self.act = nn.ReLU()
        else:
            self.act = nn.Identity()

        if use_lab is True:
            self.lab = LearnableAffineBlock(scale_value=1.0, bias_value=0.0)
        else:
            self.lab = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        x = self.bn(x)
        x = self.act(x)
        x = self.lab(x)

        return x


class LightConvBNAct(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: tuple[int, int], use_lab: bool) -> None:
        super().__init__()
        self.conv1 = Conv2dBNActivation(
            in_channels,
            out_channels,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            groups=1,
            use_act=False,
            use_lab=False,
        )
        self.conv2 = Conv2dBNActivation(
            out_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=(1, 1),
            padding=(kernel_size[0] // 2, kernel_size[1] // 2),
            groups=out_channels,
            use_act=True,
            use_lab=use_lab,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv1(x)
        x = self.conv2(x)

        return x


class EseModule(nn.Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.conv = nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.sigmoid = nn.Sigmoid()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x
        x = x.mean((2, 3), keepdim=True)
        x = self.conv(x)
        x = self.sigmoid(x)
        return torch.mul(identity, x)


class Stem(nn.Module):
    def __init__(self, channel_list: list[int]) -> None:
        super().__init__()
        layers = []
        for i in range(len(channel_list) - 1):
            layers.append(
                Conv2dBNActivation(
                    channel_list[i],
                    channel_list[i + 1],
                    kernel_size=(3, 3),
                    stride=(2, 2) if i == 0 else (1, 1),
                    padding=(1, 1),
                    groups=1,
                    use_act=True,
                    use_lab=False,
                )
            )

        self.conv = nn.Sequential(*layers)
        self.pool = nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(1, 1))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        x = self.pool(x)

        return x


class HighPerfGPUBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        mid_channels: int,
        out_channels: int,
        num_layers: int,
        kernel_size: tuple[int, int],
        residual: bool,
        light_block: bool,
        use_lab: bool,
        use_ese: bool,
        drop_path: float,
    ) -> None:
        super().__init__()
        self.residual = residual

        self.layers = nn.ModuleList()
        for i in range(num_layers):
            if light_block is True:
                self.layers.append(
                    LightConvBNAct(
                        in_channels if i == 0 else mid_channels,
                        mid_channels,
                        kernel_size=kernel_size,
                        use_lab=use_lab,
                    )
                )
            else:
                self.layers.append(
                    Conv2dBNActivation(
                        in_channels if i == 0 else mid_channels,
                        mid_channels,
                        kernel_size=kernel_size,
                        stride=(1, 1),
                        padding=(kernel_size[0] // 2, kernel_size[1] // 2),
                        groups=1,
                        use_act=True,
                        use_lab=use_lab,
                    )
                )

        # Feature aggregation
        total_chs = in_channels + num_layers * mid_channels
        if use_ese is True:
            self.aggregation = nn.Sequential(
                Conv2dBNActivation(
                    total_chs,
                    out_channels,
                    kernel_size=(1, 1),
                    stride=(1, 1),
                    padding=(0, 0),
                    groups=1,
                    use_act=True,
                    use_lab=use_lab,
                ),
                EseModule(out_channels),
            )
        else:
            self.aggregation = nn.Sequential(
                Conv2dBNActivation(
                    total_chs,
                    out_channels // 2,
                    kernel_size=(1, 1),
                    stride=(1, 1),
                    padding=(0, 0),
                    groups=1,
                    use_act=True,
                    use_lab=use_lab,
                ),
                Conv2dBNActivation(
                    out_channels // 2,
                    out_channels,
                    kernel_size=(1, 1),
                    stride=(1, 1),
                    padding=(0, 0),
                    groups=1,
                    use_act=True,
                    use_lab=use_lab,
                ),
            )

        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x
        output = [x]
        for layer in self.layers:
            x = layer(x)
            output.append(x)

        x = torch.concat(output, dim=1)
        x = self.aggregation(x)
        if self.residual is True:
            x = self.drop_path(x) + identity

        return x


class HighPerfGPUStage(nn.Module):
    def __init__(
        self,
        in_channels: int,
        mid_channels: int,
        out_channels: int,
        depth: int,
        num_layers: int,
        downsample: bool,
        stride: tuple[int, int],
        kernel_size: tuple[int, int],
        light_block: bool,
        use_lab: bool,
        use_ese: bool,
        drop_path: list[float],
    ) -> None:
        super().__init__()
        if downsample is True:
            self.downsample = Conv2dBNActivation(
                in_channels,
                in_channels,
                kernel_size=(3, 3),
                stride=stride,
                padding=(kernel_size[0] // 2, kernel_size[1] // 2),
                groups=in_channels,
                use_act=False,
                use_lab=False,
            )
        else:
            self.downsample = nn.Identity()

        blocks_list = []
        for i in range(depth):
            blocks_list.append(
                HighPerfGPUBlock(
                    in_channels if i == 0 else out_channels,
                    mid_channels,
                    out_channels,
                    num_layers,
                    residual=False if i == 0 else True,
                    kernel_size=kernel_size,
                    light_block=light_block,
                    use_lab=use_lab,
                    use_ese=use_ese,
                    drop_path=drop_path[i],
                )
            )
        self.blocks = nn.Sequential(*blocks_list)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.blocks(x)

        return x


# pylint: disable=invalid-name
class HGNet_v1(DetectorBackbone):
    # pylint: disable=too-many-locals
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        use_lab = False
        use_ese = True
        feature_dim = 2048
        stem_channels: list[int] = self.config["stem_channels"]
        stage1: list[int | bool] = self.config["stage1"]
        stage2: list[int | bool] = self.config["stage2"]
        stage3: list[int | bool] = self.config["stage3"]
        stage4: list[int | bool] = self.config["stage4"]
        drop_path_rate: float = self.config["drop_path_rate"]

        stages_cfg = [stage1, stage2, stage3, stage4]
        block_depths = [c[3] for c in stages_cfg]

        self.stem = Stem([self.input_channels] + stem_channels)

        # Stochastic depth decay rule
        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(block_depths)).split(block_depths)]

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i, stage_config in enumerate(stages_cfg):
            (in_ch, mid_ch, out_ch, blocks, downsample, light_block, k, num_layers) = stage_config
            stages[f"stage{i+1}"] = HighPerfGPUStage(
                in_channels=in_ch,
                mid_channels=mid_ch,
                out_channels=out_ch,
                depth=blocks,
                num_layers=num_layers,
                downsample=downsample,  # type: ignore[arg-type]
                stride=(2, 2),
                kernel_size=(k, k),
                light_block=light_block,  # type: ignore[arg-type]
                use_lab=use_lab,
                use_ese=use_ese,
                drop_path=dpr[i],
            )
            return_channels.append(out_ch)

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Conv2d(out_ch, feature_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
            nn.ReLU(),
            LearnableAffineBlock(scale_value=1.0, bias_value=0.0) if use_lab is True else nn.Identity(),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = feature_dim
        self.classifier = self.create_classifier()

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "hgnet_v1_tiny",
    HGNet_v1,
    config={
        "stem_channels": [48, 48, 96],
        # in, mid, out, blocks, downsample, light_block, kernel_size, num_layers
        "stage1": [96, 96, 224, 1, False, False, 3, 5],
        "stage2": [224, 128, 448, 1, True, False, 3, 5],
        "stage3": [448, 160, 512, 2, True, False, 3, 5],
        "stage4": [512, 192, 768, 1, True, False, 3, 5],
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "hgnet_v1_small",
    HGNet_v1,
    config={
        "stem_channels": [64, 64, 128],
        # in, mid, out, blocks, downsample, light_block, kernel_size, num_layers
        "stage1": [128, 128, 256, 1, False, False, 3, 6],
        "stage2": [256, 160, 512, 1, True, False, 3, 6],
        "stage3": [512, 192, 768, 2, True, False, 3, 6],
        "stage4": [768, 224, 1024, 1, True, False, 3, 6],
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "hgnet_v1_base",
    HGNet_v1,
    config={
        "stem_channels": [96, 96, 160],
        # in, mid, out, blocks, downsample, light_block, kernel_size, num_layers
        "stage1": [160, 192, 320, 1, False, False, 3, 7],
        "stage2": [320, 224, 640, 2, True, False, 3, 7],
        "stage3": [640, 256, 960, 3, True, False, 3, 7],
        "stage4": [960, 288, 1280, 2, True, False, 3, 7],
        "drop_path_rate": 0.0,
    },
)

registry.register_weights(
    "hgnet_v1_tiny_il-common",
    {
        "description": "HGNet v1 tiny model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 51.4,
                "sha256": "c8895d04c8fb28b617e5c3c3c9186979ce480912873814c2cc2bb65cdf57e0f1",
            }
        },
        "net": {"network": "hgnet_v1_tiny", "tag": "il-common"},
    },
)



import argparse
import contextlib
import logging
import math
import os
import re
import subprocess
from collections import deque
from collections.abc import Callable
from collections.abc import Generator
from collections.abc import Iterator
from datetime import datetime
from pathlib import Path
from typing import Any
from typing import Literal
from typing import Optional
from typing import Sized

import numpy as np
import torch
import torch.amp
import torch.distributed as dist
import torch.utils.data.distributed
from torchvision.ops import FrozenBatchNorm2d

from birder.conf import settings
from birder.data.transforms.classification import get_rgb_stats
from birder.data.transforms.classification import training_preset
from birder.optim import Lamb
from birder.optim import Lars
from birder.scheduler import CooldownLR

logger = logging.getLogger(__name__)

OptimizerType = Literal["sgd", "rmsprop", "adam", "adamw", "nadam", "nadamw", "lamb", "lambw", "lars"]
SchedulerType = Literal["constant", "step", "multistep", "cosine", "polynomial"]

###############################################################################
# Data Sampling
###############################################################################


class RASampler(torch.utils.data.Sampler):
    """
    Sampler that restricts data loading to a subset of the dataset for distributed,
    with repeated augmentation.
    It ensures that different each augmented version of a sample will be visible to a
    different process (GPU).
    Heavily based on 'torch.utils.data.DistributedSampler'.

    This is borrowed from the TorchVision repository:
    https://github.com/pytorch/vision/blob/v0.16.2/references/classification/sampler.py

    Based on: https://arxiv.org/pdf/2105.13343.pdf
    """

    def __init__(
        self,
        dataset: Sized,
        num_replicas: int,
        rank: int,
        shuffle: bool,
        seed: int = 0,
        repetitions: int = 3,
    ) -> None:
        super().__init__()
        self.dataset = dataset
        self.num_replicas = num_replicas
        self.rank = rank
        self.epoch = 0
        self.num_samples = int(math.ceil(len(self.dataset) * float(repetitions) / self.num_replicas))
        self.total_size = self.num_samples * self.num_replicas
        self.num_selected_samples = int(math.floor(len(self.dataset) // 256 * 256 / self.num_replicas))
        self.shuffle = shuffle
        self.seed = seed
        self.repetitions = repetitions

    def __iter__(self) -> Iterator[list[int]]:
        if self.shuffle is True:
            # Deterministically shuffle based on epoch
            g = torch.Generator()
            g.manual_seed(self.seed + self.epoch)
            indices = torch.randperm(len(self.dataset), generator=g).tolist()
        else:
            indices = list(range(len(self.dataset)))

        # Add extra samples to make it evenly divisible
        indices = [ele for ele in indices for i in range(self.repetitions)]
        indices += indices[: (self.total_size - len(indices))]
        assert len(indices) == self.total_size

        # Subsample
        indices = indices[self.rank : self.total_size : self.num_replicas]
        assert len(indices) == self.num_samples

        return iter(indices[: self.num_selected_samples])

    def __len__(self) -> int:
        return self.num_selected_samples

    def set_epoch(self, epoch: int) -> None:
        self.epoch = epoch


###############################################################################
# Model Weight Averaging
###############################################################################


class ExponentialMovingAverage(torch.optim.swa_utils.AveragedModel):
    """
    Maintains moving averages of model parameters using an exponential decay

    See: "Averaging Weights Leads to Wider Optima and Better Generalization"
    https://arxiv.org/abs/1803.05407
    """

    def __init__(self, model: torch.nn.Module, decay: float, device: torch.device) -> None:
        def ema_avg(
            avg_model_param: torch.nn.Parameter, model_param: torch.nn.Parameter, _num_averaged: int
        ) -> torch.Tensor:
            return decay * avg_model_param + (1 - decay) * model_param

        super().__init__(model, device, ema_avg, use_buffers=True)


def ema_model(args: argparse.Namespace, net: torch.nn.Module, device: torch.device) -> ExponentialMovingAverage:
    # Decay adjustment that aims to keep the decay independent of other hyper-parameters originally
    # proposed at: https://github.com/facebookresearch/pycls/blob/f8cd9627/pycls/core/net.py#L123
    #
    # total_ema_updates = (Dataset_size / n_GPUs) * epochs / (batch_size_per_gpu * EMA_steps)
    # We consider constant = Dataset_size for a given dataset/setup and omit it. Thus:
    # adjust = 1 / total_ema_updates ~= n_GPUs * batch_size_per_gpu * EMA_steps / epochs

    adjust = args.world_size * args.batch_size * args.model_ema_steps / args.epochs
    alpha = 1.0 - args.model_ema_decay
    alpha = min(1.0, alpha * adjust)
    model_ema = ExponentialMovingAverage(net, device=device, decay=1.0 - alpha)

    return model_ema


###############################################################################
# Optimizer Parameter Groups
###############################################################################


def group_by_regex(strings: list[str], pattern: str) -> list[list[str]]:
    groups: list[list[str]] = []
    current_group: list[str] = []
    current_block: Optional[tuple[str, ...]] = None

    for s in strings:
        match = re.search(pattern, s)
        if match is not None:
            block_key = match.groups()
            if block_key != current_block:
                if len(current_group) > 0:
                    groups.append(current_group)

                current_group = []
                current_block = block_key

        elif current_block is not None:
            if len(current_group) > 0:
                groups.append(current_group)

            current_group = []
            current_block = None

        current_group.append(s)

    if len(current_group) > 0:
        groups.append(current_group)

    return groups


def count_layers(model: torch.nn.Module) -> int:
    num_layers = 0
    module_stack = [model]
    visited_modules: list[int] = []
    while len(module_stack) > 0:
        skip_module = False
        module = module_stack.pop()
        if id(module) in visited_modules:
            skip_module = True

        visited_modules.append(id(module))
        parameters_found = False
        for _, _ in module.named_parameters(recurse=False):
            if skip_module is True:
                break

            parameters_found = True

        if parameters_found is True:
            num_layers += 1

        for _, child_module in reversed(list(module.named_children())):
            module_stack.append(child_module)

    return num_layers


# pylint: disable=protected-access,too-many-locals,too-many-branches
def optimizer_parameter_groups(
    model: torch.nn.Module,
    weight_decay: float,
    norm_weight_decay: Optional[float] = None,
    custom_keys_weight_decay: Optional[list[tuple[str, float]]] = None,
    layer_decay: Optional[float] = None,
    layer_decay_min_scale: Optional[float] = None,
    layer_decay_no_opt_scale: Optional[float] = None,
    bias_lr: Optional[float] = None,
    backbone_lr: Optional[float] = None,
) -> list[dict[str, Any]]:
    """
    Return parameter groups for optimizers with per-parameter group weight decay.

    This function creates parameter groups with customizable weight decay, layer-wise
    learning rate scaling, and special handling for different parameter types. It supports
    advanced optimization techniques like layer decay and custom weight decay rules.

    Referenced from https://github.com/pytorch/vision/blob/main/references/classification/utils.py and from
    https://github.com/karpathy/minGPT/blob/3ed14b2cec0dfdad3f4b2831f2b4a86d11aef150/mingpt/model.py#L136

    Layer decay optimization trick from: https://github.com/huggingface/pytorch-image-models/pull/2537/files

    Parameters
    ----------
    model
        The PyTorch model whose parameters will be grouped for optimization.
    weight_decay
        Default weight decay (L2 regularization) value applied to parameters.
    norm_weight_decay
        Weight decay value specifically for normalization layers. If None, uses weight_decay.
    custom_keys_weight_decay
        List of (parameter_name, weight_decay) tuples for applying custom weight decay
        values to specific parameters by name matching.
    layer_decay
        Layer-wise learning rate decay factor.
    layer_decay_min_scale
        Minimum learning rate scale factor when using layer decay. Prevents layers from having too small learning rates.
    layer_decay_no_opt_scale
        Learning rate scale threshold below which parameters are frozen (requires_grad=False).
    bias_lr
        Custom learning rate for bias parameters (parameters ending with '.bias').
    backbone_lr
        Custom learning rate for backbone parameters (parameters starting with 'backbone.').

    Returns
    -------
    List of parameter group dictionaries suitable for PyTorch optimizers.

    Notes
    -----
    - The function handles duplicate parameters (module aliases) by warning and skipping them
    - Layer grouping is determined by the model's `block_group_regex` attribute if available,
      otherwise layers are counted sequentially
    - Parameters with requires_grad=False are automatically skipped
    - Custom key matching supports both full parameter names and shortened names for nested modules
    """

    norm_classes = (
        torch.nn.modules.batchnorm._BatchNorm,
        torch.nn.LayerNorm,
        torch.nn.GroupNorm,
        torch.nn.RMSNorm,
        torch.nn.modules.instancenorm._InstanceNorm,
        torch.nn.LocalResponseNorm,
    )

    block_group_regex = getattr(model, "block_group_regex", None)
    if block_group_regex is not None:
        names = [n for n, _ in model.named_parameters()]
        groups = group_by_regex(names, block_group_regex)
        group_map = {item: index for index, sublist in enumerate(groups) for item in sublist}
        num_layers = len(groups)
    else:
        group_map = {}
        num_layers = count_layers(model)
        if layer_decay is not None:
            logger.warning("Assigning lr scaling (layer decay) without a block group map")

    # Build layer scale
    if layer_decay_min_scale is None:
        layer_decay_min_scale = 0.0

    layer_scales = []
    if layer_decay is not None:
        layer_max = num_layers - 1
        layer_scales = [max(layer_decay_min_scale, layer_decay ** (layer_max - i)) for i in range(num_layers)]
        logger.info(f"Layer scaling in range of {min(layer_scales)} - {max(layer_scales)} on {num_layers} layers")

    # Set weight decay and layer decay
    idx = 0
    params = []
    module_stack_with_prefix = [(model, "")]
    visited_modules = []
    while len(module_stack_with_prefix) > 0:
        skip_module = False
        (module, prefix) = module_stack_with_prefix.pop()
        if id(module) in visited_modules:
            skip_module = True

        visited_modules.append(id(module))
        parameters_found = False
        for name, p in module.named_parameters(recurse=False):
            target_name = f"{prefix}.{name}" if prefix != "" else name
            idx = group_map.get(target_name, idx)
            if skip_module is True:
                break

            parameters_found = True
            if p.requires_grad is False:
                continue
            if layer_decay is not None and layer_decay_no_opt_scale is not None:
                if layer_scales[idx] < layer_decay_no_opt_scale:
                    p.requires_grad_(False)

            is_custom_key = False
            if custom_keys_weight_decay is not None:
                for key, custom_wd in custom_keys_weight_decay:
                    target_name_for_custom_key = f"{prefix}.{name}" if prefix != "" and "." in key else name
                    if key == target_name_for_custom_key:
                        d = {
                            "params": p,
                            "weight_decay": custom_wd,
                            "lr_scale": 1.0 if layer_decay is None else layer_scales[idx],
                        }
                        if backbone_lr is not None and target_name.startswith("backbone.") is True:
                            d["lr"] = backbone_lr

                        params.append(d)
                        is_custom_key = True
                        break

            if is_custom_key is False:
                if norm_weight_decay is not None and isinstance(module, norm_classes):
                    wd = norm_weight_decay
                else:
                    wd = weight_decay

                d = {
                    "params": p,
                    "weight_decay": wd,
                    "lr_scale": 1.0 if layer_decay is None else layer_scales[idx],
                }
                if backbone_lr is not None and target_name.startswith("backbone.") is True:
                    d["lr"] = backbone_lr

                if bias_lr is not None and target_name.endswith(".bias") is True:
                    d["lr"] = bias_lr

                params.append(d)

        if parameters_found is True:
            idx += 1

        for child_name, child_module in reversed(list(module.named_children())):
            child_prefix = f"{prefix}.{child_name}" if prefix != "" else child_name
            module_stack_with_prefix.append((child_module, child_prefix))

    return params


def get_wd_custom_keys(args: argparse.Namespace) -> list[tuple[str, float]]:
    custom_keys_weight_decay = []
    if args.bias_weight_decay is not None:
        custom_keys_weight_decay.append(("bias", args.bias_weight_decay))

    if args.transformer_embedding_decay is not None:
        for key in [
            "cls_token",
            "class_token",
            "mask_token",
            "pos_embed",
            "pos_embedding",
            "pos_embed_win",
            "position_embedding",
            "relative_position_bias_table",
            "rel_pos_h",
            "rel_pos_w",
            "decoder_embed",
        ]:
            custom_keys_weight_decay.append((key, args.transformer_embedding_decay))

    return custom_keys_weight_decay


###############################################################################
# Components Setup
###############################################################################


def get_optimizer(parameters: list[dict[str, Any]], lr: float, args: argparse.Namespace) -> torch.optim.Optimizer:
    opt: OptimizerType = args.opt
    kwargs = {}
    if getattr(args, "opt_eps", None) is not None:
        kwargs["eps"] = args.opt_eps
    if getattr(args, "opt_betas", None) is not None:
        kwargs["betas"] = args.opt_betas
    if getattr(args, "opt_alpha", None) is not None:
        kwargs["alpha"] = args.opt_alpha

    if opt == "sgd":
        optimizer = torch.optim.SGD(
            parameters, lr=lr, momentum=args.momentum, nesterov=args.nesterov, weight_decay=args.wd
        )
    elif opt == "rmsprop":
        if "alpha" not in kwargs:
            kwargs["alpha"] = 0.9
        if "eps" not in kwargs:
            kwargs["eps"] = 0.0316

        optimizer = torch.optim.RMSprop(parameters, lr=lr, momentum=args.momentum, weight_decay=args.wd, **kwargs)
    elif opt == "adam":
        optimizer = torch.optim.Adam(parameters, lr=lr, weight_decay=args.wd, **kwargs)
    elif opt == "adamw":
        optimizer = torch.optim.AdamW(parameters, lr=lr, weight_decay=args.wd, **kwargs)
    elif opt == "nadam":
        optimizer = torch.optim.NAdam(parameters, lr=lr, weight_decay=args.wd, **kwargs)
    elif opt == "nadamw":
        optimizer = torch.optim.NAdam(parameters, lr=lr, weight_decay=args.wd, decoupled_weight_decay=True, **kwargs)
    elif opt == "lamb":
        optimizer = Lamb(parameters, lr=lr, weight_decay=args.wd, **kwargs)
    elif opt == "lambw":
        optimizer = Lamb(parameters, lr=lr, weight_decay=args.wd, decoupled_decay=True, **kwargs)
    elif opt == "lars":
        optimizer = Lars(
            parameters, lr=lr, momentum=args.momentum, nesterov=args.nesterov, weight_decay=args.wd, **kwargs
        )
    else:
        raise ValueError("Unknown optimizer")

    return optimizer


def get_scheduler(
    optimizer: torch.optim.Optimizer, iters_per_epoch: int, args: argparse.Namespace
) -> torch.optim.lr_scheduler.LRScheduler:
    # At first, we translate everything into "steps"
    begin_step = 0
    if args.resume_epoch is not None:
        begin_step = args.resume_epoch * iters_per_epoch

    steps = args.epochs * iters_per_epoch

    if args.warmup_epochs is not None:
        warmup_steps = args.warmup_epochs * iters_per_epoch
    elif args.warmup_iters is not None:
        warmup_steps = args.warmup_iters
    else:
        warmup_steps = 0

    if args.cooldown_epochs is not None:
        cooldown_steps = args.cooldown_epochs * iters_per_epoch
    elif args.cooldown_iters is not None:
        cooldown_steps = args.cooldown_iters
    else:
        cooldown_steps = 0

    # Warmup steps is given in absolute number from 0
    remaining_warmup = max(0, warmup_steps - begin_step)

    # Cooldown steps is given in absolute number from the end
    remaining_cooldown = min(cooldown_steps, steps - begin_step)

    main_steps = steps - begin_step - remaining_warmup - remaining_cooldown - 1

    logger.debug(f"Using {iters_per_epoch} steps per epoch")
    logger.debug(
        f"Scheduler {args.lr_scheduler} set for {steps} steps of which {warmup_steps} "
        f"are warmup and {cooldown_steps} cooldown"
    )
    logger.debug(
        f"Currently starting from step {begin_step} with {remaining_warmup} remaining warmup steps "
        f"and {remaining_cooldown} remaining cooldown steps"
    )
    logger.debug(f"Main scheduler has {main_steps} steps")

    if args.lr_scheduler == "constant":
        main_scheduler = torch.optim.lr_scheduler.ConstantLR(optimizer, factor=1.0, total_iters=1)
    elif args.lr_scheduler == "step":
        main_scheduler = torch.optim.lr_scheduler.StepLR(
            optimizer, step_size=args.lr_step_size, gamma=args.lr_step_gamma
        )
    elif args.lr_scheduler == "multistep":
        main_scheduler = torch.optim.lr_scheduler.MultiStepLR(
            optimizer, milestones=args.lr_steps, gamma=args.lr_step_gamma
        )
    elif args.lr_scheduler == "cosine":
        main_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
            optimizer, T_max=main_steps, eta_min=args.lr_cosine_min
        )
    elif args.lr_scheduler == "polynomial":
        main_scheduler = torch.optim.lr_scheduler.PolynomialLR(optimizer, total_iters=main_steps, power=args.lr_power)
    else:
        raise ValueError("Unknown learning rate scheduler")

    # Handle warmup and cooldown
    if warmup_steps > 0 or cooldown_steps > 0:
        schedulers = []
        milestones = []
        if warmup_steps > 0:
            warmup_lr_scheduler = torch.optim.lr_scheduler.LinearLR(
                optimizer,
                start_factor=args.lr_warmup_decay if remaining_warmup > 0 else 1.0,
                total_iters=remaining_warmup,
            )
            schedulers.append(warmup_lr_scheduler)
            milestones.append(remaining_warmup)

        schedulers.append(main_scheduler)

        if cooldown_steps > 0:
            cooldown_lr_scheduler = CooldownLR(optimizer, total_steps=remaining_cooldown)
            schedulers.append(cooldown_lr_scheduler)
            milestones.append(remaining_warmup + main_steps + 1)

        scheduler = torch.optim.lr_scheduler.SequentialLR(
            optimizer,
            schedulers=schedulers,
            milestones=milestones,
        )

    else:
        scheduler = main_scheduler

    return scheduler


def get_amp_scaler(amp: bool, amp_dtype_str: str) -> tuple[Optional[torch.amp.GradScaler], Optional[torch.dtype]]:
    if amp is True:
        scaler = torch.amp.GradScaler("cuda")
        amp_dtype = getattr(torch, amp_dtype_str)

    else:
        scaler = None
        amp_dtype = None

    return (scaler, amp_dtype)


def get_samplers(
    args: argparse.Namespace, training_dataset: torch.utils.data.Dataset, validation_dataset: torch.utils.data.Dataset
) -> torch.utils.data.Sampler:
    if args.distributed is True:
        if args.ra_sampler is True:
            train_sampler = RASampler(
                training_dataset,
                num_replicas=args.world_size,
                rank=args.rank,
                shuffle=True,
                repetitions=args.ra_reps,
            )

        else:
            train_sampler = torch.utils.data.distributed.DistributedSampler(training_dataset, shuffle=True)

        validation_sampler = torch.utils.data.distributed.DistributedSampler(validation_dataset, shuffle=False)

    else:
        train_sampler = torch.utils.data.RandomSampler(training_dataset)
        validation_sampler = torch.utils.data.SequentialSampler(validation_dataset)

    return (train_sampler, validation_sampler)


def get_training_transform(args: argparse.Namespace) -> Callable[..., torch.Tensor]:
    return training_preset(
        args.size,
        args.aug_type,
        args.aug_level,
        get_rgb_stats(args.rgb_mode),
        args.resize_min_scale,
        args.re_prob,
        args.use_grayscale,
        args.ra_num_ops,
        args.ra_magnitude,
        args.augmix_severity,
        args.simple_crop,
    )


###############################################################################
# Metrics Tracking and Reporting
###############################################################################


def to_tensor(x: torch.Tensor | float, device: Optional[torch.device] = None) -> torch.Tensor:
    if isinstance(x, torch.Tensor):
        return x.to(device) if device is not None else x

    return torch.tensor(x, device=device)


class SmoothedValue:
    # Adapted from: https://github.com/facebookresearch/capi/blob/main/utils.py

    def __init__(self, window_size: int = 32) -> None:
        self.window_size = window_size
        self.deque: deque[torch.Tensor | float] = deque(maxlen=window_size)
        self.total: torch.Tensor | float = 0.0
        self.count: int = 0

    def update(self, value: torch.Tensor | float, n: int = 1) -> None:
        self.deque.append(value)
        self.count += n
        self.total += value * n

    def synchronize_between_processes(self, device: torch.device, op: dist.ReduceOp = dist.ReduceOp.AVG) -> None:
        if is_dist_available_and_initialized() is False:
            return

        count = to_tensor(self.count).to(device=device, dtype=torch.float64).reshape(1)
        total = to_tensor(self.total).to(device=device, dtype=torch.float64).reshape(1)
        tensor_deque = torch.tensor(list(self.deque), dtype=torch.float64, device=device)
        t = torch.concat([count, total, tensor_deque], dim=0)
        dist.barrier()
        dist.all_reduce(t, op=op)
        self.count = int(t[0].cpu().item())
        self.total = t[1]
        self.deque = deque(list(t[2:]), maxlen=self.window_size)

    @property
    def median(self) -> float:
        d = torch.tensor(list(self.deque))
        return d.median().cpu().item()  # type: ignore[no-any-return]

    @property
    def avg(self) -> float:
        d = torch.tensor(list(self.deque), dtype=torch.float32)
        return d.mean().cpu().item()  # type: ignore[no-any-return]

    @property
    def global_avg(self) -> float:
        return to_tensor(self.total, torch.device("cpu")).item() / self.count  # type: ignore[no-any-return]

    @property
    def max(self) -> float:
        return torch.tensor(self.deque).max().cpu().item()  # type: ignore[no-any-return]

    @property
    def value(self) -> float:
        v = self.deque[-1]
        return to_tensor(v, torch.device("cpu")).item()  # type: ignore[no-any-return]


def accuracy(y_true: torch.Tensor, y_pred: torch.Tensor) -> float:
    if y_pred.dim() > 1 and y_pred.size(1) > 1:
        y_pred = y_pred.argmax(dim=1)

    y_true = y_true.flatten()
    y_pred = y_pred.flatten()

    return (y_true == y_pred).float().mean().item()  # type: ignore[no-any-return]


###############################################################################
# Distributed Training Utilities
###############################################################################


def init_distributed_mode(args: argparse.Namespace) -> None:
    if "RANK" in os.environ and "WORLD_SIZE" in os.environ:
        # torch.distributed.run, torchrun
        logger.debug("Detected PyTorch distributed environment (torchrun / torch.distributed.launch)")
        args.rank = int(os.environ["RANK"])
        args.world_size = int(os.environ["WORLD_SIZE"])
        args.local_rank = int(os.environ["LOCAL_RANK"])
    elif "SLURM_PROCID" in os.environ:
        # Slurm
        logger.debug("Detected Slurm distributed environment")
        args.rank = int(os.environ["SLURM_PROCID"])
        args.world_size = int(os.environ["SLURM_NTASKS"])
        args.local_rank = int(os.environ["SLURM_LOCALID"])
    elif "OMPI_COMM_WORLD_RANK" in os.environ:
        # Open MPI
        logger.debug("Detected Open MPI distributed environment")
        args.rank = int(os.environ["OMPI_COMM_WORLD_RANK"])
        args.world_size = int(os.environ["OMPI_COMM_WORLD_SIZE"])
        args.local_rank = int(os.environ["OMPI_COMM_WORLD_LOCAL_RANK"])
    elif args.world_size > 1:
        # Other, by CLI
        logger.debug("Using distributed mode with CLI-provided parameters")
    else:
        logger.debug("No distributed environment detected, running in single-process mode")
        args.rank = 0
        args.distributed = False
        if args.local_rank is None:
            args.local_rank = 0

        torch.cuda.set_device(args.local_rank)
        return

    torch.cuda.set_device(args.local_rank)

    args.distributed = args.world_size > 1
    if args.distributed is False:
        logger.info("World size is 1, not using distributed mode")
        return

    # Initialize process group
    logger.debug(f"Using {args.dist_backend} backend for distributed training")
    logger.info(f"Distributed init (rank {args.rank}), total {args.world_size}: {args.dist_url}")
    dist.init_process_group(
        backend=args.dist_backend, init_method=args.dist_url, world_size=args.world_size, rank=args.rank
    )

    # Synchronize all processes
    logger.debug("Synchronizing all processes...")
    dist.barrier(device_ids=[args.rank])
    if is_local_primary(args) is False:
        logger.debug(f"Non-primary process (rank {args.rank}), disabling logging")
        disable_print()
        logging.disable(logging.WARNING)


def shutdown_distributed_mode(args: argparse.Namespace) -> None:
    if args.distributed is True:
        dist.destroy_process_group()


def disable_print() -> None:
    import builtins as __builtin__  # pylint: disable=import-outside-toplevel

    builtin_print = __builtin__.print

    def print(*args, **kwargs):  # type: ignore  # pylint: disable=redefined-builtin
        force = kwargs.pop("force", False)
        if force is True:
            builtin_print(*args, **kwargs)

    __builtin__.print = print


def is_dist_available_and_initialized() -> bool:
    if dist.is_available() is False:
        return False
    if dist.is_initialized() is False:
        return False

    return True


def reduce_across_processes(value: torch.Tensor | float, device: torch.device, op: dist.ReduceOp) -> float:
    if is_dist_available_and_initialized() is False:
        return value

    value = to_tensor(value, device)
    dist.barrier()
    dist.all_reduce(value, op)

    return value.item()  # type: ignore[no-any-return]


def get_world_size() -> int:
    if is_dist_available_and_initialized() is False:
        return 1

    return dist.get_world_size()  # type: ignore[no-any-return]


def is_global_primary(args: argparse.Namespace) -> bool:
    return args.rank == 0  # type: ignore[no-any-return]


def is_local_primary(args: argparse.Namespace) -> bool:
    if is_dist_available_and_initialized() is False:
        return True

    return args.local_rank == 0  # type: ignore[no-any-return]


###############################################################################
# Utility Functions
###############################################################################


def cosine_scheduler(
    base_value: float,
    final_value: float,
    epochs: int,
    warmup_epochs: float,
    iter_per_epoch: int,
    start_warmup_value: float = 0.0,
) -> list[float]:
    """
    Create a learning rate schedule with linear warmup followed by cosine annealing

    The schedule consists of two phases:
    1. Linear warmup: linearly increases from start_warmup_value to base_value
    2. Cosine annealing: follows cosine decay from base_value to final_value

    Parameters
    ----------
    base_value
        Peak value after warmup phase (start of cosine decay).
    final_value
        Final value at the end of training.
    epochs
        Total number of training epochs.
    warmup_epochs
        Number of epochs for warmup phase. Can be fractional (e.g., 0.5 for half epoch).
    iter_per_epoch
        Number of iterations per epoch.
    start_warmup_value
        Starting value for warmup phase.

    Returns
    -------
    List of scheduler values for each iteration. Length equals epochs * iter_per_epoch.
    """

    if warmup_epochs >= epochs:
        raise ValueError(f"warmup_epochs ({warmup_epochs}) must be less than epochs ({epochs})")
    if epochs <= 0 or iter_per_epoch <= 0:
        raise ValueError("epochs and iter_per_epoch must be positive")
    if warmup_epochs < 0:
        raise ValueError("warmup_epochs must be non-negative")

    total_iterations = epochs * iter_per_epoch
    warmup_iterations = int(warmup_epochs * iter_per_epoch)
    cosine_iterations = total_iterations - warmup_iterations

    warmup_schedule = np.linspace(start_warmup_value, base_value, warmup_iterations, endpoint=False)
    iters = np.arange(cosine_iterations)
    cosine_schedule = final_value + 0.5 * (base_value - final_value) * (1 + np.cos(np.pi * iters / len(iters)))

    schedule = np.concatenate((warmup_schedule, cosine_schedule))
    assert len(schedule) == total_iterations

    return schedule.tolist()  # type: ignore[no-any-return]


def scale_lr(args: argparse.Namespace) -> float:
    lr: float = args.lr
    if args.lr_scale is not None:
        ratio = args.batch_size * args.grad_accum_steps * args.world_size / args.lr_scale
        if args.lr_scale_type == "sqrt":
            ratio = ratio**0.5

        lr = lr * ratio
        logger.info(f"Adjusted learning rate to: {lr}")

    return lr


def log_git_info() -> None:
    cwd = os.path.dirname(os.path.abspath(__file__))

    def _run(command: list[str]) -> str:
        try:
            return (
                subprocess.check_output(command, cwd=cwd, stderr=subprocess.DEVNULL)  # nosec # only hard-coded args
                .decode("ascii")
                .strip()
            )
        except Exception:  # pylint: disable=broad-exception-caught
            return "N/A"

    sha = _run(["git", "rev-parse", "HEAD"])
    status = "unclean" if _run(["git", "diff-index", "HEAD"]) else "clean"
    branch = _run(["git", "rev-parse", "--abbrev-ref", "HEAD"])

    logger.info(f"[GIT] sha: {sha}, status: {status}, branch: {branch}")


def training_log_name(network: str, device: torch.device) -> str:
    timestamp = datetime.now().replace(microsecond=0)
    if is_dist_available_and_initialized() is True:
        posix_ts = timestamp.timestamp()
        posix_ts_t = torch.tensor(posix_ts, dtype=torch.float64, device=device)
        dist.broadcast(posix_ts_t, src=0, async_op=False)
        posix_ts = posix_ts_t.item()
        timestamp = datetime.fromtimestamp(posix_ts)

    iso_timestamp = timestamp.strftime("%Y-%m-%dT%H%M%S")
    return f"{network}__{iso_timestamp}"


def training_log_path(network_name: str, device: torch.device, experiment: Optional[str]) -> Path:
    log_name = training_log_name(network_name, device)
    if experiment is not None:
        log_dir = settings.TRAINING_RUNS_PATH.joinpath(experiment)
        log_dir.mkdir(exist_ok=True)
    else:
        log_dir = settings.TRAINING_RUNS_PATH

    return log_dir.joinpath(log_name)


def setup_file_logging(log_file_path: str | Path) -> logging.Handler:
    file_handler = logging.FileHandler(log_file_path)
    formatter = logging.Formatter(
        fmt="{message}",
        style="{",
    )
    file_handler.setFormatter(formatter)
    file_handler.setLevel(settings.LOG_LEVEL)

    birder_logger = logging.getLogger("birder")
    birder_logger.addHandler(file_handler)

    return file_handler


@contextlib.contextmanager
def single_handler_logging(
    target_logger: logging.Logger, target_handler: logging.Handler, enabled: bool = True
) -> Generator[logging.Logger, None, None]:
    """
    Context manager to temporarily use only a specific handler for logging.

    Parameters
    ----------
    target_logger
        The logger to modify.
    target_handler
        The handler to keep (all others will be temporarily removed).
    enabled
        If False, acts as a no-op and yields the original logger unchanged.

    Yields
    ------
    The logger object, either modified to use only the target handler
    (if enabled=True) or unchanged (if enabled=False).

    Examples
    --------
    >>> with single_handler_logging(logger, file_handler, enabled=file_only_logging) as log:
    ...     log.info("This only goes to the target handler")
    """

    if enabled is False:
        yield target_logger
        return

    # Store original state
    original_handlers = list(target_logger.handlers)
    original_propagate = target_logger.propagate

    # Remove all handlers
    for handler in original_handlers:
        target_logger.removeHandler(handler)

    target_logger.addHandler(target_handler)

    # Disable propagation to parent loggers
    target_logger.propagate = False

    try:
        yield target_logger
    finally:
        # Restore all handlers in the original order
        target_logger.removeHandler(target_handler)
        for handler in original_handlers:
            target_logger.addHandler(handler)

        target_logger.propagate = original_propagate


def get_grad_norm(parameters: Iterator[torch.Tensor], norm_type: float = 2.0) -> float:
    filtered_parameters = list(filter(lambda p: p.grad is not None, parameters))
    norm_type = float(norm_type)
    total_norm: float = 0.0
    for p in filtered_parameters:
        param_norm = p.grad.data.norm(norm_type)
        total_norm += param_norm.item() ** norm_type

    total_norm = total_norm ** (1.0 / norm_type)

    return total_norm


def freeze_batchnorm2d(module: torch.nn.Module) -> torch.nn.Module:
    """
    Referenced from https://github.com/huggingface/pytorch-image-models/blob/main/timm/layers/norm_act.py#L251
    """

    res = module
    if isinstance(module, (torch.nn.modules.batchnorm.BatchNorm2d, torch.nn.modules.batchnorm.SyncBatchNorm)):
        res = FrozenBatchNorm2d(module.num_features)
        res.num_features = module.num_features
        res.affine = module.affine
        if module.affine is True:
            res.weight.data = module.weight.data.clone().detach()
            res.bias.data = module.bias.data.clone().detach()

        res.running_mean.data = module.running_mean.data
        res.running_var.data = module.running_var.data
        res.eps = module.eps
    else:
        for name, child in module.named_children():
            new_child = freeze_batchnorm2d(child)
            if new_child is not child:
                res.add_module(name, new_child)

    return res


def replace_module(
    module: torch.nn.Module,
    from_module: type[torch.nn.Module] | tuple[type[torch.nn.Module], ...],
    to_module: type[torch.nn.Module],
) -> torch.nn.Module:
    res = module
    if isinstance(module, from_module):
        res = to_module()
    else:
        for name, child in module.named_children():
            new_child = replace_module(child, from_module, to_module)
            if new_child is not child:
                res.add_module(name, new_child)

    return res



"""
SimCLR, mostly generated by Claude Opus 4

Paper "A Simple Framework for Contrastive Learning of Visual Representations",
https://arxiv.org/abs/2002.05709
"""

from typing import Any
from typing import Optional

import torch
import torch.distributed as dist
import torch.nn.functional as F
from torch import nn

from birder.common import training_utils
from birder.net.base import BaseNet
from birder.net.ssl.base import SSLBaseNet


# pylint: disable=abstract-method,arguments-differ
class FullGatherLayer(torch.autograd.Function):
    @staticmethod
    def forward(_ctx, x):  # type: ignore
        if training_utils.is_dist_available_and_initialized() is False:
            return (x,)

        output = [torch.zeros_like(x) for _ in range(dist.get_world_size())]
        dist.all_gather(output, x)
        return tuple(output)

    @staticmethod
    def backward(_ctx, *grads):  # type: ignore
        all_gradients = torch.stack(grads)
        if training_utils.is_dist_available_and_initialized() is False:
            return all_gradients

        dist.all_reduce(all_gradients)
        return all_gradients[dist.get_rank()]


class SimCLR(SSLBaseNet):
    def __init__(
        self,
        backbone: BaseNet,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(backbone, config=config, size=size)
        assert self.config is not None, "must set config"

        projection_dim: int = self.config["projection_dim"]
        projection_hidden_dim: int = self.config["projection_hidden_dim"]
        temperature: float = self.config["temperature"]

        self.temperature = temperature
        self.projection_head = nn.Sequential(
            nn.Linear(self.backbone.embedding_size, projection_hidden_dim),
            nn.ReLU(),
            nn.Linear(projection_hidden_dim, projection_dim, bias=False),
        )

    def _contrastive_loss(self, features: torch.Tensor) -> torch.Tensor:
        """
        Compute NT-Xent loss (Normalized Temperature-scaled Cross Entropy Loss)
        """

        device = features.device
        batch_size = features.size(0) // 2

        features = F.normalize(features, dim=1)
        similarity_matrix = torch.matmul(features, features.t())
        mask = torch.eye(batch_size * 2, dtype=torch.bool, device=device)

        similarity_matrix = similarity_matrix[~mask].reshape(batch_size * 2, batch_size * 2 - 1)
        similarity_matrix = similarity_matrix / self.temperature

        # Create labels: for sample i, positive is at i+batch_size (and vice versa)
        labels = torch.concat([torch.arange(batch_size) + batch_size - 1, torch.arange(batch_size)], dim=0)
        labels = labels.to(device)

        # Compute loss
        loss = F.cross_entropy(similarity_matrix, labels, reduction="mean")

        return loss

    # pylint: disable=arguments-differ
    def forward(self, x1: torch.Tensor, x2: torch.Tensor) -> torch.Tensor:  # type: ignore[override]
        h1 = self.backbone.embedding(x1)
        h2 = self.backbone.embedding(x2)

        z1 = self.projection_head(h1)
        z2 = self.projection_head(h2)

        features = torch.concat([z1, z2], dim=0)
        features = torch.concat(FullGatherLayer.apply(features), dim=0)

        loss = self._contrastive_loss(features)

        return loss



import argparse
import logging
from typing import Any

import torch
from torch.utils.data import DataLoader

import birder
from birder.common import cli
from birder.conf import settings
from birder.data.datasets.directory import make_image_dataset
from birder.inference.data_parallel import InferenceDataParallel

logger = logging.getLogger(__name__)


def evaluate(args: argparse.Namespace) -> None:
    if args.gpu is True:
        device = torch.device("cuda")
    elif args.mps is True:
        device = torch.device("mps")
    else:
        device = torch.device("cpu")

    if args.parallel is True and torch.cuda.device_count() > 1:
        logger.info(f"Using {torch.cuda.device_count()} {device} devices")
    else:
        if args.gpu_id is not None:
            torch.cuda.set_device(args.gpu_id)

        logger.info(f"Using device {device}")

    if args.fast_matmul is True or args.amp is True:
        torch.set_float32_matmul_precision("high")

    model_dtype: torch.dtype = getattr(torch, args.model_dtype)
    amp_dtype: torch.dtype = getattr(torch, args.amp_dtype)
    model_list = birder.list_pretrained_models(args.filter)
    for model_name in model_list:
        (net, (class_to_idx, signature, rgb_stats, *_)) = birder.load_pretrained_model(
            model_name, inference=True, device=device, dtype=model_dtype
        )
        if args.parallel is True and torch.cuda.device_count() > 1:
            net = InferenceDataParallel(net, output_device="cpu", compile_replicas=args.compile)
        if args.compile is True:
            net = torch.compile(net)

        if args.size is None:
            size = birder.get_size_from_signature(signature)
        else:
            size = args.size

        transform = birder.classification_transform(size, rgb_stats, args.center_crop)
        dataset = make_image_dataset(args.data_path, class_to_idx, transforms=transform)
        num_samples = len(dataset)
        inference_loader = DataLoader(dataset, batch_size=args.batch_size, num_workers=8)
        with torch.inference_mode():
            results = birder.evaluate_classification(
                device,
                net,
                inference_loader,
                class_to_idx,
                args.tta,
                model_dtype,
                args.amp,
                amp_dtype,
                num_samples=num_samples,
                sparse=args.save_sparse_results,
            )

        logger.info(f"{model_name}: accuracy={results.accuracy:.4f}")
        base_output_path = (
            f"{args.dir}/{model_name}_{len(class_to_idx)}_{size[0]}px_crop{args.center_crop}_{num_samples}"
        )
        if args.save_sparse_results is True:
            results_file_suffix = "_sparse.csv"
        else:
            results_file_suffix = ".csv"

        results.save(f"{base_output_path}{results_file_suffix}")


def get_args_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description="evaluate pretrained models on specified dataset",
        epilog=(
            "Usage example:\n"
            "python -m birder.scripts.evaluate --filter '*il-all*' --fast-matmul --gpu data/validation_il-all_packed\n"
            "python -m birder.scripts.evaluate --amp --compile --gpu --gpu-id 1 data/testing\n"
            "python -m birder.scripts.evaluate --filter '*inat21*' --amp --compile --gpu "
            "--parallel ~/Datasets/inat2021/val\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    parser.add_argument("--filter", type=str, help="models to evaluate (fnmatch type filter)")
    parser.add_argument("--compile", default=False, action="store_true", help="enable compilation")
    parser.add_argument(
        "--model-dtype",
        type=str,
        choices=["float32", "float16", "bfloat16"],
        default="float32",
        help="model dtype to use",
    )
    parser.add_argument(
        "--amp", default=False, action="store_true", help="use torch.amp.autocast for mixed precision inference"
    )
    parser.add_argument(
        "--amp-dtype",
        type=str,
        choices=["float16", "bfloat16"],
        default="float16",
        help="whether to use float16 or bfloat16 for mixed precision",
    )
    parser.add_argument(
        "--fast-matmul", default=False, action="store_true", help="use fast matrix multiplication (affects precision)"
    )
    parser.add_argument("--tta", default=False, action="store_true", help="test time augmentation (oversampling)")
    parser.add_argument(
        "--size", type=int, nargs="+", metavar=("H", "W"), help="image size for inference (defaults to model signature)"
    )
    parser.add_argument("--batch-size", type=int, default=64, metavar="N", help="the batch size")
    parser.add_argument("--center-crop", type=float, default=1.0, help="center crop ratio to use during inference")
    parser.add_argument(
        "--dir", type=str, default="evaluate", help="place all outputs in a sub-directory (relative to results)"
    )
    parser.add_argument("--gpu", default=False, action="store_true", help="use gpu")
    parser.add_argument("--gpu-id", type=int, metavar="ID", help="gpu id to use")
    parser.add_argument("--mps", default=False, action="store_true", help="use mps (Metal Performance Shaders) device")
    parser.add_argument("--parallel", default=False, action="store_true", help="use multiple gpus")
    parser.add_argument(
        "--save-sparse-results",
        default=False,
        action="store_true",
        help="save results object in memory-efficient sparse format (only top-k probabilities)",
    )
    parser.add_argument("data_path", nargs="+", help="data files path (directories and files)")

    return parser


def validate_args(args: argparse.Namespace) -> None:
    assert args.amp is False or args.model_dtype == "float32"
    args.size = cli.parse_size(args.size)


def args_from_dict(**kwargs: Any) -> argparse.Namespace:
    parser = get_args_parser()
    parser.set_defaults(**kwargs)
    args = parser.parse_args([])
    validate_args(args)

    return args


def main() -> None:
    parser = get_args_parser()
    args = parser.parse_args()
    validate_args(args)

    if settings.RESULTS_DIR.joinpath(args.dir).exists() is False:
        logger.info(f"Creating {settings.RESULTS_DIR.joinpath(args.dir)} directory...")
        settings.RESULTS_DIR.joinpath(args.dir).mkdir(parents=True)

    evaluate(args)


if __name__ == "__main__":
    logger = logging.getLogger(getattr(__spec__, "name", __name__))
    main()



"""
RepGhost, adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/repghost.py

Paper "RepGhost: A Hardware-Efficient Ghost Module via Re-parameterization", https://arxiv.org/abs/2211.06088
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import SqueezeExcitation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import make_divisible


class RepGhostModule(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        dw_size: int,
        stride: int,
        relu: bool,
        reparameterized: bool,
    ) -> None:
        super().__init__()
        self.reparameterized = reparameterized
        self.primary_conv = nn.Sequential(
            nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size=(kernel_size, kernel_size),
                stride=(stride, stride),
                padding=(kernel_size // 2, kernel_size // 2),
                bias=False,
            ),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True) if relu else nn.Identity(),
        )

        if reparameterized is True:
            self.fusion_bn = None
        else:
            self.fusion_bn = nn.BatchNorm2d(out_channels)

        self.cheap_operation = nn.Sequential(
            nn.Conv2d(
                out_channels,
                out_channels,
                kernel_size=(dw_size, dw_size),
                stride=(1, 1),
                padding=(dw_size // 2, dw_size // 2),
                groups=out_channels,
                bias=False,
            ),
            nn.BatchNorm2d(out_channels),
        )
        self.relu = nn.ReLU() if relu else nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x1 = self.primary_conv(x)
        x2 = self.cheap_operation(x1)
        if self.fusion_bn is not None:
            x2 = x2 + self.fusion_bn(x1)

        return self.relu(x2)

    def reparameterize(self) -> None:
        if self.reparameterized is True:
            return

        (kernel, bias) = self._get_kernel_bias()
        self.cheap_operation = nn.Conv2d(
            in_channels=self.cheap_operation[0].in_channels,
            out_channels=self.cheap_operation[0].out_channels,
            kernel_size=self.cheap_operation[0].kernel_size,
            padding=self.cheap_operation[0].padding,
            dilation=self.cheap_operation[0].dilation,
            groups=self.cheap_operation[0].groups,
            bias=True,
        )

        self.cheap_operation.weight.data = kernel
        self.cheap_operation.bias.data = bias

        del self.fusion_bn
        self.fusion_bn = None
        self.reparameterized = True

    def _get_kernel_bias(self) -> tuple[torch.Tensor, torch.Tensor]:
        (kernel, bias) = self._fuse_bn_tensor(self.cheap_operation[0], self.cheap_operation[1])
        if self.fusion_bn is not None:
            (kernel1x1, bias_bn) = self._fuse_bn_tensor(nn.Identity(), self.fusion_bn, kernel.shape[0])
            kernel += F.pad(kernel1x1, [1, 1, 1, 1])
            bias += bias_bn

        return (kernel, bias)

    def _fuse_bn_tensor(
        self, conv: nn.Module, bn: nn.Module, in_channels: Optional[int] = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        if in_channels is None:
            in_channels = bn.running_mean.shape[0]

        if isinstance(conv, nn.Conv2d):
            kernel = conv.weight
            assert conv.bias is None
        else:
            assert isinstance(conv, nn.Identity)
            device = bn.weight.device
            kernel = torch.ones(in_channels, 1, 1, 1, device=device)

        running_mean = bn.running_mean
        running_var = bn.running_var
        gamma = bn.weight
        beta = bn.bias
        eps = bn.eps

        std = (running_var + eps).sqrt()
        t = (gamma / std).reshape(-1, 1, 1, 1)

        return (kernel * t, beta - running_mean * gamma / std)


class RepGhostBottleneck(nn.Module):
    def __init__(
        self,
        in_channels: int,
        mid_channels: int,
        out_channels: int,
        dw_size: tuple[int, int],
        stride: tuple[int, int],
        se_ratio: float,
        reparameterized: bool,
    ) -> None:
        super().__init__()

        # Point-wise expansion
        self.ghost1 = RepGhostModule(
            in_channels, mid_channels, kernel_size=1, dw_size=3, stride=1, relu=True, reparameterized=reparameterized
        )

        # Depth-wise convolution
        if stride[0] > 1 or stride[1] > 1:
            self.conv_dw = nn.Conv2d(
                mid_channels,
                mid_channels,
                kernel_size=dw_size,
                stride=stride,
                padding=(dw_size[0] // 2, dw_size[1] // 2),
                groups=mid_channels,
                bias=False,
            )
            self.bn_dw = nn.BatchNorm2d(mid_channels)
        else:
            self.conv_dw = nn.Identity()
            self.bn_dw = nn.Identity()

        # Squeeze-and-excitation
        if se_ratio > 0:
            self.se = SqueezeExcitation(
                mid_channels, make_divisible(int(mid_channels * se_ratio), 4), activation=nn.Hardsigmoid
            )
        else:
            self.se = nn.Identity()

        # Point-wise linear projection
        self.ghost2 = RepGhostModule(
            mid_channels, out_channels, kernel_size=1, dw_size=3, stride=1, relu=False, reparameterized=reparameterized
        )

        # Shortcut
        if in_channels == out_channels and stride[0] == 1 and stride[1] == 1:
            self.shortcut = nn.Identity()
        else:
            self.shortcut = nn.Sequential(
                nn.Conv2d(
                    in_channels,
                    in_channels,
                    kernel_size=dw_size,
                    stride=stride,
                    padding=(dw_size[0] // 2, dw_size[1] // 2),
                    groups=in_channels,
                    bias=False,
                ),
                nn.BatchNorm2d(in_channels),
                nn.Conv2d(in_channels, out_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
                nn.BatchNorm2d(out_channels),
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        shortcut = x
        x = self.ghost1(x)
        x = self.conv_dw(x)
        x = self.bn_dw(x)
        x = self.se(x)
        x = self.ghost2(x)
        x = x + self.shortcut(shortcut)

        return x


class RepGhost(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        width: float = self.config["width"]

        self.reparameterized = False
        block_config: list[list[tuple[int, int, int, float, int]]] = [
            # kernel, expansion, channels, se ratio, stride
            # Stage 1
            [(3, 8, 16, 0, 1)],
            # Stage 2
            [(3, 24, 24, 0, 2)],
            [(3, 36, 24, 0, 1)],
            # Stage 3
            [(5, 36, 40, 0.25, 2)],
            [(5, 60, 40, 0.25, 1)],
            # Stage 4
            [(3, 120, 80, 0, 2)],
            [
                (3, 100, 80, 0, 1),
                (3, 120, 80, 0, 1),
                (3, 120, 80, 0, 1),
                (3, 240, 112, 0.25, 1),
                (3, 336, 112, 0.25, 1),
            ],
            # Stage 5
            [(5, 336, 160, 0.25, 2)],
            [
                (5, 480, 160, 0, 1),
                (5, 480, 160, 0.25, 1),
                (5, 480, 160, 0, 1),
                (5, 480, 160, 0.25, 1),
            ],
        ]

        stem_channels = make_divisible(16 * width, 4)
        self.stem = Conv2dNormActivation(
            self.input_channels, stem_channels, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), bias=False
        )

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        i = 0
        layers: list[nn.Module] = []
        prev_channels = stem_channels
        for cfg in block_config:
            for k, exp_size, c, se_ratio, s in cfg:
                if s > 1:
                    stages[f"stage{i}"] = nn.Sequential(*layers)
                    return_channels.append(prev_channels)
                    layers = []
                    i += 1

                out_channels = make_divisible(c * width, 4)
                mid_channels = make_divisible(exp_size * width, 4)
                layers.append(
                    RepGhostBottleneck(
                        prev_channels,
                        mid_channels,
                        out_channels,
                        (k, k),
                        (s, s),
                        se_ratio=se_ratio,
                        reparameterized=self.reparameterized,
                    )
                )
                prev_channels = out_channels

        out_channels = make_divisible(exp_size * width * 2, 4)
        layers.append(
            Conv2dNormActivation(prev_channels, out_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        )
        stages[f"stage{i}"] = nn.Sequential(*layers)
        return_channels.append(out_channels)

        self.body = nn.Sequential(stages)
        prev_channels = out_channels
        out_channels = 1280
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Conv2d(prev_channels, out_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=True),
            nn.ReLU(inplace=True),
            nn.Flatten(1),
            nn.Dropout(p=0.2),
        )
        self.return_channels = return_channels[1:5]
        self.embedding_size = out_channels
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def reparameterize_model(self) -> None:
        for module in self.modules():
            if hasattr(module, "reparameterize") is True:
                module.reparameterize()

        self.reparameterized = True


registry.register_model_config("repghost_0_5", RepGhost, config={"width": 0.5})
registry.register_model_config("repghost_1_0", RepGhost, config={"width": 1.0})
registry.register_model_config("repghost_1_3", RepGhost, config={"width": 1.3})
registry.register_model_config("repghost_1_5", RepGhost, config={"width": 1.5})

registry.register_weights(
    "repghost_1_0_il-common",
    {
        "description": "RepGhost 1.0x model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 12.8,
                "sha256": "37e211ec65c752ad79bbbaacea277f7d683d0b0f69d954a7ca7af46b9a1260e6",
            }
        },
        "net": {"network": "repghost_1_0", "tag": "il-common"},
    },
)
registry.register_weights(
    "repghost_1_0_il-common_reparameterized",
    {
        "description": "RepGhost 1.0x (reparameterized) model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 12.6,
                "sha256": "e003e0498d63428305c10f879a0e2b999604795d417f07ea0da35ea925f794f5",
            }
        },
        "net": {"network": "repghost_1_0", "tag": "il-common_reparameterized", "reparameterized": True},
    },
)



"""
SwiftFormer, adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/swiftformer.py

Paper "SwiftFormer: Efficient Additive Attention for Transformer-based Real-time Mobile Vision Applications",
https://arxiv.org/abs/2303.15446
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import StochasticDepth

from birder.layers import LayerScale2d
from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class PatchEmbed(nn.Module):
    def __init__(
        self,
        in_channels: int,
        embed_dim: int,
        patch_size: tuple[int, int],
        patch_stride: tuple[int, int],
        padding: tuple[int, int],
    ) -> None:
        super().__init__()
        self.proj = nn.Conv2d(in_channels, embed_dim, kernel_size=patch_size, stride=patch_stride, padding=padding)
        self.norm = nn.BatchNorm2d(embed_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.proj(x)
        x = self.norm(x)

        return x


class ConvEncoder(nn.Module):
    def __init__(
        self, dim: int, hidden_dim: int, kernel_size: tuple[int, int], drop_path: float, use_layer_scale: bool
    ) -> None:
        super().__init__()
        self.dw_conv = nn.Conv2d(
            dim, dim, kernel_size, stride=(1, 1), padding=(kernel_size[0] // 2, kernel_size[1] // 2), groups=dim
        )
        self.norm = nn.BatchNorm2d(dim)
        self.pw_conv1 = nn.Conv2d(dim, hidden_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.act = nn.GELU()
        self.pw_conv2 = nn.Conv2d(hidden_dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.drop_path = StochasticDepth(drop_path, "row")
        if use_layer_scale is True:
            self.layer_scale = LayerScale2d(dim, init_value=1.0)
        else:
            self.layer_scale = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        shortcut = x
        x = self.dw_conv(x)
        x = self.norm(x)
        x = self.pw_conv1(x)
        x = self.act(x)
        x = self.pw_conv2(x)
        x = self.layer_scale(x)
        x = shortcut + self.drop_path(x)

        return x


class ConvMLP(nn.Module):
    def __init__(self, in_features: int, hidden_features: int, drop: float) -> None:
        super().__init__()
        self.norm1 = nn.BatchNorm2d(in_features)
        self.fc1 = nn.Conv2d(in_features, hidden_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.act = nn.GELU()
        self.fc2 = nn.Conv2d(hidden_features, in_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.drop = nn.Dropout(drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.norm1(x)
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)

        return x


class EfficientAdditiveAttention(nn.Module):
    def __init__(self, in_dims: int, token_dim: int, num_heads: int) -> None:
        super().__init__()
        self.scale_factor = token_dim**-0.5
        self.to_query = nn.Linear(in_dims, token_dim * num_heads)
        self.to_key = nn.Linear(in_dims, token_dim * num_heads)

        self.w_g = nn.Parameter(torch.randn(token_dim * num_heads, 1))

        self.proj = nn.Linear(token_dim * num_heads, token_dim * num_heads)
        self.final = nn.Linear(token_dim * num_heads, token_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, _, H, W) = x.size()
        x = x.flatten(2).permute(0, 2, 1)

        query = F.normalize(self.to_query(x), dim=-1)
        key = F.normalize(self.to_key(x), dim=-1)

        attn = F.normalize(query @ self.w_g * self.scale_factor, dim=1)
        attn = torch.sum(attn * query, dim=1, keepdim=True)

        out = self.proj(attn * key) + query
        out = self.final(out).permute(0, 2, 1).reshape(B, -1, H, W)

        return out


class LocalRepresentation(nn.Module):
    def __init__(self, dim: int, kernel_size: tuple[int, int], drop_path: float, use_layer_scale: bool) -> None:
        super().__init__()
        self.dw_conv = nn.Conv2d(
            dim, dim, kernel_size, stride=(1, 1), padding=(kernel_size[0] // 2, kernel_size[1] // 2), groups=dim
        )
        self.norm = nn.BatchNorm2d(dim)
        self.pw_conv1 = nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.act = nn.GELU()
        self.pw_conv2 = nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.drop_path = StochasticDepth(drop_path, "row")
        if use_layer_scale is True:
            self.layer_scale = LayerScale2d(dim, init_value=1.0)
        else:
            self.layer_scale = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        shortcut = x
        x = self.dw_conv(x)
        x = self.norm(x)
        x = self.pw_conv1(x)
        x = self.act(x)
        x = self.pw_conv2(x)
        x = self.layer_scale(x)
        x = shortcut + self.drop_path(x)

        return x


class SwiftFormerBlock(nn.Module):
    def __init__(self, dim: int, mlp_ratio: float, drop_rate: float, drop_path: float, use_layer_scale: bool) -> None:
        super().__init__()
        self.local_representation = LocalRepresentation(
            dim=dim, kernel_size=(3, 3), drop_path=0.0, use_layer_scale=use_layer_scale
        )
        self.attn = EfficientAdditiveAttention(dim, token_dim=dim, num_heads=1)
        self.linear = ConvMLP(dim, hidden_features=int(dim * mlp_ratio), drop=drop_rate)
        self.drop_path = StochasticDepth(drop_path, "row")
        if use_layer_scale is True:
            self.layer_scale_1 = LayerScale2d(dim, init_value=1e-5)
            self.layer_scale_2 = LayerScale2d(dim, init_value=1e-5)
        else:
            self.layer_scale_1 = nn.Identity()
            self.layer_scale_2 = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.local_representation(x)
        x = x + self.drop_path(self.layer_scale_1(self.attn(x)))
        x = x + self.drop_path(self.layer_scale_2(self.linear(x)))

        return x


class SwiftFormerStage(nn.Module):
    def __init__(
        self,
        dim: int,
        dim_out: int,
        index: int,
        layers: list[int],
        mlp_ratio: float,
        drop_rate: float,
        drop_path_rate: float,
        use_layer_scale: bool,
        downsample: bool,
    ) -> None:
        super().__init__()
        if downsample is True:
            self.downsample = PatchEmbed(dim, dim_out, patch_size=(3, 3), patch_stride=(2, 2), padding=(1, 1))
        else:
            assert dim == dim_out
            self.downsample = nn.Identity()

        blocks = []
        for block_idx in range(layers[index]):
            block_dpr = drop_path_rate * (block_idx + sum(layers[:index])) / (sum(layers) - 1)
            if layers[index] - block_idx <= 1:
                blocks.append(
                    SwiftFormerBlock(
                        dim_out,
                        mlp_ratio=mlp_ratio,
                        drop_rate=drop_rate,
                        drop_path=block_dpr,
                        use_layer_scale=use_layer_scale,
                    )
                )
            else:
                blocks.append(
                    ConvEncoder(
                        dim=dim_out,
                        hidden_dim=int(mlp_ratio * dim_out),
                        kernel_size=(3, 3),
                        drop_path=block_dpr,
                        use_layer_scale=use_layer_scale,
                    )
                )

        self.blocks = nn.Sequential(*blocks)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.blocks(x)

        return x


class SwiftFormer(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        layers: list[int] = self.config["layers"]
        embed_dims: list[int] = self.config["embed_dims"]
        drop_path_rate: float = self.config["drop_path_rate"]

        self.stem = nn.Sequential(
            nn.Conv2d(self.input_channels, embed_dims[0] // 2, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
            nn.BatchNorm2d(embed_dims[0] // 2),
            nn.ReLU(),
            nn.Conv2d(embed_dims[0] // 2, embed_dims[0], kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
            nn.BatchNorm2d(embed_dims[0]),
            nn.ReLU(),
        )

        prev_dim = embed_dims[0]
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(len(layers)):
            stages[f"stage{i+1}"] = SwiftFormerStage(
                prev_dim,
                embed_dims[i],
                index=i,
                layers=layers,
                mlp_ratio=4.0,
                drop_rate=0.0,
                drop_path_rate=drop_path_rate,
                use_layer_scale=True,
                downsample=i > 0,
            )
            return_channels.append(embed_dims[i])
            prev_dim = embed_dims[i]

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.BatchNorm2d(embed_dims[-1]),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = embed_dims[-1]
        self.dist_classifier = self.create_classifier()
        self.classifier = self.create_classifier()
        self.distillation_output = False

        # Weight initialization
        for m in self.modules():
            if isinstance(m, (nn.Conv2d, nn.Linear)):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, nn.BatchNorm2d):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

    def reset_classifier(self, num_classes: int) -> None:
        self.num_classes = num_classes
        self.dist_classifier = self.create_classifier()
        self.classifier = self.create_classifier()

    def freeze(self, freeze_classifier: bool = True, unfreeze_features: bool = False) -> None:
        for param in self.parameters():
            param.requires_grad = False

        if freeze_classifier is False:
            for param in self.classifier.parameters():
                param.requires_grad = True

            for param in self.dist_classifier.parameters():
                param.requires_grad = True

        if unfreeze_features is True:
            for param in self.features.parameters():
                param.requires_grad = True

    def transform_to_backbone(self) -> None:
        self.features = nn.Identity()
        self.classifier = nn.Identity()
        self.dist_classifier = nn.Identity()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def set_distillation_output(self, enable: bool = True) -> None:
        self.distillation_output = enable

    def classify(self, x: torch.Tensor) -> torch.Tensor:
        x_cls = self.classifier(x)
        x_dist = self.dist_classifier(x)

        if self.training is True and self.distillation_output is True:
            x = torch.stack([x_cls, x_dist], dim=1)
        else:
            # Classifier "token" as an average of both tokens (during normal training or inference)
            x = (x_cls + x_dist) / 2

        return x


registry.register_model_config(
    "swiftformer_xs",
    SwiftFormer,
    config={"layers": [3, 3, 6, 4], "embed_dims": [48, 56, 112, 220], "drop_path_rate": 0.0},
)
registry.register_model_config(
    "swiftformer_s",
    SwiftFormer,
    config={"layers": [3, 3, 9, 6], "embed_dims": [48, 64, 168, 224], "drop_path_rate": 0.0},
)
registry.register_model_config(
    "swiftformer_l1",
    SwiftFormer,
    config={"layers": [4, 3, 10, 5], "embed_dims": [48, 96, 192, 384], "drop_path_rate": 0.0},
)
registry.register_model_config(
    "swiftformer_l3",
    SwiftFormer,
    config={"layers": [4, 4, 12, 6], "embed_dims": [64, 128, 320, 512], "drop_path_rate": 0.0},
)

registry.register_weights(
    "swiftformer_xs_il-common",
    {
        "description": "SwiftFormer model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 12.3,
                "sha256": "4053b9726f888c05afb98a6d13463da35573be1a66ca347d17208a2b3f9baba6",
            }
        },
        "net": {"network": "swiftformer_xs", "tag": "il-common"},
    },
)



"""
MogaNet, adapted from
https://github.com/Westlake-AI/MogaNet/blob/main/models/moganet.py

Paper "MogaNet: Multi-order Gated Aggregation Network",
https://arxiv.org/abs/2211.03295
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType


class ElementScale(nn.Module):
    def __init__(self, embed_dims: int, init_value: float) -> None:
        super().__init__()
        self.scale = nn.Parameter(init_value * torch.ones((1, embed_dims, 1, 1)), requires_grad=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * self.scale


class ChannelAggregationFFN(nn.Module):
    def __init__(self, embed_dims: int, feedforward_channels: int, kernel_size: int, ffn_drop: float) -> None:
        super().__init__()
        self.fc1 = nn.Conv2d(
            in_channels=embed_dims, out_channels=feedforward_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)
        )
        self.dwconv = nn.Conv2d(
            feedforward_channels,
            feedforward_channels,
            kernel_size=(kernel_size, kernel_size),
            stride=(1, 1),
            padding=(kernel_size // 2, kernel_size // 2),
            groups=feedforward_channels,
        )
        self.act = nn.GELU()
        self.drop = nn.Dropout(ffn_drop)

        self.decompose = nn.Conv2d(feedforward_channels, 1, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))  # C -> 1
        self.sigma = ElementScale(feedforward_channels, init_value=1e-5)
        self.decompose_act = nn.GELU()
        self.fc2 = nn.Conv2d(feedforward_channels, embed_dims, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Proj 1
        x = self.fc1(x)
        x = self.dwconv(x)
        x = self.act(x)
        x = self.drop(x)

        # Proj 2
        x = x + self.sigma(x - self.decompose_act(self.decompose(x)))
        x = self.fc2(x)
        x = self.drop(x)

        return x


class MultiOrderDWConv(nn.Module):
    def __init__(self, embed_dims: int, dw_dilation: list[int], channel_split: list[int]) -> None:
        super().__init__()

        split_ratio = [i / sum(channel_split) for i in channel_split]
        self.embed_dims_1 = int(split_ratio[1] * embed_dims)
        self.embed_dims_2 = int(split_ratio[2] * embed_dims)
        self.embed_dims_0 = embed_dims - self.embed_dims_1 - self.embed_dims_2
        self.embed_dims = embed_dims
        assert len(dw_dilation) == len(channel_split) == 3
        assert 1 <= min(dw_dilation) and max(dw_dilation) <= 3
        assert embed_dims % sum(channel_split) == 0

        # Padding: dilation * (kernel_size[0] - 1) // 2, dilation * (kernel_size[1] - 1) // 2
        self.dw_conv0 = nn.Conv2d(
            self.embed_dims,
            self.embed_dims,
            kernel_size=(5, 5),
            stride=(1, 1),
            padding=((1 + 4 * dw_dilation[0]) // 2, (1 + 4 * dw_dilation[0]) // 2),
            dilation=dw_dilation[0],
            groups=self.embed_dims,
        )
        self.dw_conv1 = nn.Conv2d(
            self.embed_dims_1,
            self.embed_dims_1,
            kernel_size=(5, 5),
            stride=(1, 1),
            padding=((1 + 4 * dw_dilation[1]) // 2, (1 + 4 * dw_dilation[1]) // 2),
            dilation=dw_dilation[1],
            groups=self.embed_dims_1,
        )
        self.dw_conv2 = nn.Conv2d(
            self.embed_dims_2,
            self.embed_dims_2,
            kernel_size=(7, 7),
            stride=(1, 1),
            padding=((1 + 6 * dw_dilation[2]) // 2, (1 + 6 * dw_dilation[2]) // 2),
            dilation=dw_dilation[2],
            groups=self.embed_dims_2,
        )
        self.pw_conv = nn.Conv2d(embed_dims, embed_dims, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_0 = self.dw_conv0(x)
        x_1 = self.dw_conv1(x_0[:, self.embed_dims_0 : self.embed_dims_0 + self.embed_dims_1, ...])
        x_2 = self.dw_conv2(x_0[:, self.embed_dims - self.embed_dims_2 :, ...])

        x = torch.concat([x_0[:, : self.embed_dims_0, ...], x_1, x_2], dim=1)
        x = self.pw_conv(x)

        return x


class MultiOrderGatedAggregation(nn.Module):
    def __init__(self, embed_dims: int, attn_dw_dilation: list[int], attn_channel_split: list[int]) -> None:
        super().__init__()

        self.proj_1 = nn.Conv2d(
            in_channels=embed_dims, out_channels=embed_dims, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)
        )
        self.gate = nn.Conv2d(
            in_channels=embed_dims, out_channels=embed_dims, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)
        )
        self.value = MultiOrderDWConv(
            embed_dims=embed_dims, dw_dilation=attn_dw_dilation, channel_split=attn_channel_split
        )
        self.proj_2 = nn.Conv2d(
            in_channels=embed_dims, out_channels=embed_dims, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)
        )

        self.act_value = nn.SiLU()
        self.act_gate = nn.SiLU()
        self.sigma = ElementScale(embed_dims, init_value=1e-5)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        shortcut = x

        x = self.proj_1(x)
        x_d = F.adaptive_avg_pool2d(x, output_size=(1, 1))
        x = x + self.sigma(x - x_d)
        x = self.act_value(x)

        g = self.gate(x)
        v = self.value(x)

        x = self.proj_2(self.act_gate(g) * self.act_gate(v))
        x = x + shortcut

        return x


class MogaBlock(nn.Module):
    def __init__(
        self,
        embed_dims: int,
        ffn_ratio: float,
        drop_rate: float,
        drop_path_rate: float,
        init_value: float,
        attn_dw_dilation: list[int],
        attn_channel_split: list[int],
    ) -> None:
        super().__init__()

        # Spatial attention
        self.norm1 = nn.BatchNorm2d(embed_dims, eps=1e-5)
        self.attn = MultiOrderGatedAggregation(
            embed_dims, attn_dw_dilation=attn_dw_dilation, attn_channel_split=attn_channel_split
        )
        self.layer_scale_1 = nn.Parameter(init_value * torch.ones((1, embed_dims, 1, 1)), requires_grad=True)

        # Channel MLP
        self.norm2 = nn.BatchNorm2d(embed_dims, eps=1e-5)
        mlp_hidden_dim = int(embed_dims * ffn_ratio)
        self.mlp = ChannelAggregationFFN(
            embed_dims=embed_dims,
            feedforward_channels=mlp_hidden_dim,
            kernel_size=3,
            ffn_drop=drop_rate,
        )
        self.layer_scale_2 = nn.Parameter(init_value * torch.ones((1, embed_dims, 1, 1)), requires_grad=True)

        self.drop_path = StochasticDepth(drop_path_rate, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Spatial
        identity = x
        x = self.layer_scale_1 * self.attn(self.norm1(x))
        x = identity + self.drop_path(x)

        # Channel
        identity = x
        x = self.layer_scale_2 * self.mlp(self.norm2(x))
        x = identity + self.drop_path(x)

        return x


class MogaNet(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"body\.stage(\d+)\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        init_value = 1e-5
        drop_rate = 0.0
        attn_dw_dilation = [1, 2, 3]
        attn_channel_split = [1, 3, 4]
        ffn_ratios = [8, 8, 4, 4]
        embed_dims: list[int] = self.config["embed_dims"]
        depths: list[int] = self.config["depths"]
        drop_path_rate: float = self.config["drop_path_rate"]

        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels,
                embed_dims[0] // 2,
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                activation_layer=nn.GELU,
                inplace=None,
            ),
            Conv2dNormActivation(
                embed_dims[0] // 2,
                embed_dims[0],
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                activation_layer=None,
            ),
        )

        total_depth = sum(depths)
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, total_depth)]

        cur_block_idx = 0
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i, depth in enumerate(depths):
            blocks = []
            if i > 0:
                blocks.append(
                    Conv2dNormActivation(
                        embed_dims[i - 1],
                        embed_dims[i],
                        kernel_size=(3, 3),
                        stride=(2, 2),
                        padding=(1, 1),
                        activation_layer=None,
                    )
                )

            for j in range(depth):
                blocks.append(
                    MogaBlock(
                        embed_dims=embed_dims[i],
                        ffn_ratio=ffn_ratios[i],
                        drop_rate=drop_rate,
                        drop_path_rate=dpr[cur_block_idx + j],
                        init_value=init_value,
                        attn_dw_dilation=attn_dw_dilation,
                        attn_channel_split=attn_channel_split,
                    )
                )
            cur_block_idx += depth
            blocks.append(nn.BatchNorm2d(embed_dims[i]))
            stages[f"stage{i+1}"] = nn.Sequential(*blocks)
            return_channels.append(embed_dims[i])

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = embed_dims[-1]
        self.classifier = self.create_classifier()

        self.stem_stride = 4
        self.stem_width = embed_dims[0]
        self.encoding_size = embed_dims[-1]

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(x, mask, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token)
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "moganet_xt",
    MogaNet,
    config={"embed_dims": [32, 64, 96, 192], "depths": [3, 3, 10, 2], "drop_path_rate": 0.05},
)
registry.register_model_config(
    "moganet_t",
    MogaNet,
    config={"embed_dims": [32, 64, 128, 256], "depths": [3, 3, 12, 2], "drop_path_rate": 0.1},
)
registry.register_model_config(
    "moganet_s",
    MogaNet,
    config={"embed_dims": [64, 128, 320, 512], "depths": [2, 3, 12, 2], "drop_path_rate": 0.1},
)
registry.register_model_config(
    "moganet_b",
    MogaNet,
    config={"embed_dims": [64, 160, 320, 512], "depths": [4, 6, 22, 3], "drop_path_rate": 0.2},
)
registry.register_model_config(
    "moganet_l",
    MogaNet,
    config={"embed_dims": [64, 160, 320, 640], "depths": [4, 6, 44, 4], "drop_path_rate": 0.3},
)
registry.register_model_config(
    "moganet_xl",
    MogaNet,
    config={"embed_dims": [96, 192, 480, 960], "depths": [6, 6, 44, 4], "drop_path_rate": 0.4},
)

registry.register_weights(
    "moganet_xt_il-common",
    {
        "description": "MogaNet X-Tiny model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 11.1,
                "sha256": "2b353981631650be749bc5d51b4af3aafe36a562be3f5cb0a2d91cec383a3253",
            }
        },
        "net": {"network": "moganet_xt", "tag": "il-common"},
    },
)



import argparse
import logging
from collections.abc import Callable
from typing import Any

import torch

from birder.common import cli
from birder.common import fs_ops
from birder.common import lib
from birder.data.transforms.classification import inference_preset
from birder.introspection import AttentionRolloutInterpreter
from birder.introspection import GradCamInterpreter
from birder.introspection import GuidedBackpropInterpreter
from birder.net.base import BaseNet

logger = logging.getLogger(__name__)


def _nhwc_reshape_transform(tensor: torch.Tensor) -> torch.Tensor:
    return tensor.permute(0, 3, 1, 2).contiguous()


def show_attn_rollout(
    args: argparse.Namespace,
    net: BaseNet,
    _class_to_idx: dict[str, int],
    transform: Callable[..., torch.Tensor],
    device: torch.device,
) -> None:
    ar = AttentionRolloutInterpreter(net, device, transform, args.attn_layer_name, args.discard_ratio, args.head_fusion)
    result = ar.interpret(args.image_path)
    result.show()


def show_guided_backprop(
    args: argparse.Namespace,
    net: BaseNet,
    class_to_idx: dict[str, int],
    transform: Callable[..., torch.Tensor],
    device: torch.device,
) -> None:
    if args.target is not None:
        target = class_to_idx[args.target]
    else:
        target = None

    guided_bp = GuidedBackpropInterpreter(net, device, transform)
    result = guided_bp.interpret(args.image_path, target_class=target)
    result.show()


def show_grad_cam(
    args: argparse.Namespace,
    net: BaseNet,
    class_to_idx: dict[str, int],
    transform: Callable[..., torch.Tensor],
    device: torch.device,
) -> None:
    if args.channels_last is True:
        reshape_transform = _nhwc_reshape_transform
    else:
        reshape_transform = None

    block = getattr(net, args.block_name)
    target_layer = block[args.layer_num]

    if args.target is not None:
        target = class_to_idx[args.target]
    else:
        target = None

    grad_cam = GradCamInterpreter(net, device, transform, target_layer, reshape_transform=reshape_transform)
    result = grad_cam.interpret(args.image_path, target_class=target)
    result.show()


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "introspection",
        allow_abbrev=False,
        help="computer vision introspection and explainability",
        description="computer vision introspection and explainability",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools introspection --method gradcam --network efficientnet_v2_m "
            "--epoch 200 'data/training/European goldfinch/000300.jpeg'\n"
            "python -m birder.tools introspection --method gradcam -n resnest_50 --epoch 300 "
            "data/index5.jpeg --target 'Grey heron'\n"
            "python -m birder.tools introspection --method guided-backprop -n efficientnet_v2_s "
            "-e 0 'data/training/European goldfinch/000300.jpeg'\n"
            "python -m birder.tools introspection --method gradcam -n swin_transformer_v1_b -e 85 --layer-num -4 "
            "--channels-last data/training/Fieldfare/000002.jpeg\n"
            "python -m birder.tools introspection --method attn-rollout -n vit_reg4_b16 -t mim -e 100 "
            " data/validation/Bluethroat/000013.jpeg\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument(
        "--method", type=str, choices=["gradcam", "guided-backprop", "attn-rollout"], help="introspection method"
    )
    subparser.add_argument(
        "-n", "--network", type=str, required=True, help="the neural network to use (i.e. resnet_v2)"
    )
    subparser.add_argument("-e", "--epoch", type=int, metavar="N", help="model checkpoint to load")
    subparser.add_argument("-t", "--tag", type=str, help="model tag (from the training phase)")
    subparser.add_argument(
        "-r", "--reparameterized", default=False, action="store_true", help="load reparameterized model"
    )
    subparser.add_argument("--gpu", default=False, action="store_true", help="use gpu")
    subparser.add_argument("--gpu-id", type=int, metavar="ID", help="gpu id to use")
    subparser.add_argument(
        "--size", type=int, nargs="+", metavar=("H", "W"), help="image size for inference (defaults to model signature)"
    )
    subparser.add_argument(
        "--target", type=str, help="target class, leave empty to use predicted class (gradcam and guided-backprop only)"
    )
    subparser.add_argument("--block-name", type=str, default="body", help="target block (gradcam only)")
    subparser.add_argument(
        "--layer-num", type=int, default=-1, help="target layer, index for target block (gradcam only)"
    )
    subparser.add_argument(
        "--channels-last", default=False, action="store_true", help="channels last model, like swin (gradcam only)"
    )
    subparser.add_argument(
        "--attn-layer-name", type=str, default="self_attention", help="attention layer name (attn-rollout only)"
    )
    subparser.add_argument(
        "--head-fusion",
        type=str,
        choices=["mean", "max", "min"],
        default="max",
        help="how to fuse the attention heads for attention rollout (attn-rollout only)",
    )
    subparser.add_argument(
        "--discard-ratio",
        type=float,
        default=0.9,
        help="how many of the lowest attention paths should be discarded (attn-rollout only)",
    )
    subparser.add_argument("image_path", type=str, help="input image path")
    subparser.set_defaults(func=main)


def main(args: argparse.Namespace) -> None:
    args.size = cli.parse_size(args.size)
    if args.gpu is True:
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")

    if args.gpu_id is not None:
        torch.cuda.set_device(args.gpu_id)

    logger.info(f"Using device {device}")

    (net, model_info) = fs_ops.load_model(
        device,
        args.network,
        tag=args.tag,
        epoch=args.epoch,
        new_size=args.size,
        inference=False,
        reparameterized=args.reparameterized,
    )
    if args.size is None:
        args.size = lib.get_size_from_signature(model_info.signature)

    transform = inference_preset(args.size, model_info.rgb_stats, 1.0)

    if args.method == "gradcam":
        show_grad_cam(args, net, model_info.class_to_idx, transform, device)
    elif args.method == "guided-backprop":
        show_guided_backprop(args, net, model_info.class_to_idx, transform, device)
    elif args.method == "attn-rollout":
        show_attn_rollout(args, net, model_info.class_to_idx, transform, device)



"""
DINO v2, adapted from
https://github.com/facebookresearch/dinov2/tree/main/dinov2

Paper "DINOv2: Learning Robust Visual Features without Supervision", https://arxiv.org/abs/2304.07193
"""

# Reference license: Apache-2.0

from typing import Any
from typing import Optional

import torch
import torch.distributed as dist
import torch.nn.functional as F
from torch import nn

from birder.common import training_utils
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.ssl.base import SSLBaseNet


class DINOHead(nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        use_bn: bool,
        num_layers: int,
        hidden_dim: int,
        bottleneck_dim: int,
    ) -> None:
        super().__init__()
        if num_layers == 1:
            self.mlp = nn.Linear(in_dim, bottleneck_dim)
        else:
            layers = [nn.Linear(in_dim, hidden_dim)]
            if use_bn is True:
                layers.append(nn.BatchNorm1d(hidden_dim))

            layers.append(nn.GELU())
            for _ in range(num_layers - 2):
                layers.append(nn.Linear(hidden_dim, hidden_dim))
                if use_bn is True:
                    layers.append(nn.BatchNorm1d(hidden_dim))

                layers.append(nn.GELU())

            layers.append(nn.Linear(hidden_dim, bottleneck_dim))
            self.mlp = nn.Sequential(*layers)

        self.last_layer = nn.utils.parametrizations.weight_norm(nn.Linear(bottleneck_dim, out_dim, bias=False))
        self.last_layer.parametrizations.weight.original0.data.fill_(1)

        # Weight initialization
        for m in self.mlp.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.mlp(x)
        eps = 1e-6 if x.dtype == torch.float16 else 1e-12
        x = F.normalize(x, dim=-1, p=2, eps=eps)
        x = self.last_layer(x)

        return x


class DINOLoss(nn.Module):
    def __init__(self, out_dim: int, student_temp: float, center_momentum: float) -> None:
        super().__init__()
        self.student_temp = student_temp
        self.center_momentum = center_momentum
        self.center = nn.Buffer(torch.zeros(1, out_dim))

        self.updated = True
        self.reduce_handle: Any = None
        self.len_teacher_output: Optional[int] = None
        self.async_batch_center: Optional[torch.Tensor] = None

    def forward(
        self, student_output_list: list[torch.Tensor], teacher_out_softmax_centered_list: list[torch.Tensor]
    ) -> float:
        total_loss = 0.0
        for s in student_output_list:
            lsm = F.log_softmax(s / self.student_temp, dim=-1)
            for t in teacher_out_softmax_centered_list:
                loss = torch.sum(t * lsm, dim=-1)
                total_loss -= loss.mean()

        return total_loss

    @torch.no_grad()  # type: ignore[misc]
    def softmax_center_teacher(self, teacher_output: torch.Tensor, teacher_temp: float) -> torch.Tensor:
        self.apply_center_update()
        return F.softmax((teacher_output - self.center) / teacher_temp, dim=-1)

    @torch.no_grad()  # type: ignore[misc]
    def sinkhorn_knopp_teacher(
        self, teacher_output: torch.Tensor, teacher_temp: float, n_iterations: int = 3
    ) -> torch.Tensor:
        world_size = training_utils.get_world_size()

        teacher_output = teacher_output.float()
        q = torch.exp(teacher_output / teacher_temp).t()  # Q is K-by-B for consistency with notations from the paper
        B = q.size(1) * world_size  # Number of samples to assign
        k = q.size(0)  # How many prototypes

        sum_q = torch.sum(q)
        if training_utils.is_dist_available_and_initialized() is True:
            dist.all_reduce(sum_q)

        q /= sum_q

        for _ in range(n_iterations):
            # Normalize each row: total weight per prototype must be 1/K
            sum_of_rows = torch.sum(q, dim=1, keepdim=True)
            if training_utils.is_dist_available_and_initialized() is True:
                dist.all_reduce(sum_of_rows)

            q /= sum_of_rows
            q /= k

            # Normalize each column: total weight per sample must be 1/B
            q /= torch.sum(q, dim=0, keepdim=True)
            q /= B

        q *= B  # The columns must sum to 1 so that Q is an assignment

        return q.t()

    @torch.no_grad()  # type: ignore[misc]
    def update_center(self, teacher_output: torch.Tensor) -> None:
        self.reduce_center_update(teacher_output)

    @torch.no_grad()  # type: ignore[misc]
    def reduce_center_update(self, teacher_output: torch.Tensor) -> None:
        self.updated = False
        self.len_teacher_output = len(teacher_output)
        self.async_batch_center = torch.sum(teacher_output, dim=0, keepdim=True)
        if dist.is_initialized():
            self.reduce_handle = dist.all_reduce(self.async_batch_center, async_op=True)

    @torch.no_grad()  # type: ignore[misc]
    def apply_center_update(self) -> None:
        if self.updated is False:
            world_size = training_utils.get_world_size()
            if self.reduce_handle is not None:
                self.reduce_handle.wait()

            _t = self.async_batch_center / (self.len_teacher_output * world_size)  # type: ignore[operator]
            self.center = self.center * self.center_momentum + _t * (1 - self.center_momentum)
            self.updated = True


# pylint: disable=invalid-name
class iBOTPatchLoss(nn.Module):
    def __init__(self, patch_out_dim: int, student_temp: float, center_momentum: float) -> None:
        super().__init__()
        self.student_temp = student_temp
        self.center_momentum = center_momentum
        self.center = nn.Buffer(torch.zeros(1, 1, patch_out_dim))
        self.updated = True
        self.reduce_handle: Any = None
        self.len_teacher_patch_tokens: Optional[int] = None
        self.async_batch_center: Optional[torch.Tensor] = None

    def forward(
        self, student_patch_tokens: torch.Tensor, teacher_patch_tokens: torch.Tensor, student_masks_flat: torch.Tensor
    ) -> torch.Tensor:
        loss = torch.sum(teacher_patch_tokens * F.log_softmax(student_patch_tokens / self.student_temp, dim=-1), dim=-1)
        loss = torch.sum(loss * student_masks_flat.float(), dim=-1) / student_masks_flat.sum(dim=-1).clamp(min=1.0)

        return -loss.mean()

    def forward_masked(
        self,
        student_patch_tokens_masked: torch.Tensor,
        teacher_patch_tokens_masked: torch.Tensor,
        student_masks_flat: torch.Tensor,
        masks_weight: torch.Tensor,
        n_masked_patches: Optional[int] = None,
    ) -> torch.Tensor:
        t = teacher_patch_tokens_masked
        s = student_patch_tokens_masked
        loss = torch.sum(t * F.log_softmax(s / self.student_temp, dim=-1), dim=-1)
        if n_masked_patches is not None:
            loss = loss[:n_masked_patches]

        loss = loss * masks_weight

        return -loss.sum() / student_masks_flat.size(0)

    @torch.no_grad()  # type: ignore[misc]
    def softmax_center_teacher(self, teacher_patch_tokens: torch.Tensor, teacher_temp: float) -> torch.Tensor:
        self.apply_center_update()
        return F.softmax((teacher_patch_tokens - self.center) / teacher_temp, dim=-1)

    @torch.no_grad()  # type: ignore[misc]
    def sinkhorn_knopp_teacher(
        self,
        teacher_output: torch.Tensor,
        teacher_temp: float,
        n_masked_patches_tensor: torch.Tensor,
        n_iterations: int = 3,
    ) -> torch.Tensor:
        teacher_output = teacher_output.float()
        q = torch.exp(teacher_output / teacher_temp).t()  # Q is K-by-B for consistency with notations from the paper
        B = n_masked_patches_tensor
        if training_utils.is_dist_available_and_initialized() is True:
            dist.all_reduce(B)

        K = q.size(0)  # How many prototypes

        sum_q = torch.sum(q)
        if training_utils.is_dist_available_and_initialized() is True:
            dist.all_reduce(sum_q)

        q /= sum_q

        for _ in range(n_iterations):
            # Normalize each row: total weight per prototype must be 1/K
            sum_of_rows = torch.sum(q, dim=1, keepdim=True)
            if training_utils.is_dist_available_and_initialized() is True:
                dist.all_reduce(sum_of_rows)

            q /= sum_of_rows
            q /= K

            # Normalize each column: total weight per sample must be 1/B
            q /= torch.sum(q, dim=0, keepdim=True)
            q /= B

        q *= B  # The columns must sum to 1 so that Q is an assignment

        return q.t()

    @torch.no_grad()  # type: ignore[misc]
    def update_center(self, teacher_patch_tokens: torch.Tensor) -> None:
        self.reduce_center_update(teacher_patch_tokens)

    @torch.no_grad()  # type: ignore[misc]
    def reduce_center_update(self, teacher_patch_tokens: torch.Tensor) -> None:
        self.updated = False
        self.len_teacher_patch_tokens = len(teacher_patch_tokens)
        self.async_batch_center = torch.sum(teacher_patch_tokens.mean(1), dim=0, keepdim=True)
        if training_utils.is_dist_available_and_initialized() is True:
            self.reduce_handle = dist.all_reduce(self.async_batch_center, async_op=True)

    @torch.no_grad()  # type: ignore[misc]
    def apply_center_update(self) -> None:
        if self.updated is False:
            world_size = training_utils.get_world_size()
            if self.reduce_handle is not None:
                self.reduce_handle.wait()

            _t = self.async_batch_center / (self.len_teacher_patch_tokens * world_size)  # type: ignore[operator]
            self.center = self.center * self.center_momentum + _t * (1 - self.center_momentum)
            self.updated = True


class KoLeoLoss(nn.Module):
    """
    Kozachenko-Leonenko entropic loss regularizer from:
    Spreading vectors for similarity search - https://arxiv.org/abs/1806.03198
    """

    def __init__(self) -> None:
        super().__init__()
        self.pdist = nn.PairwiseDistance(2, eps=1e-8)

    def pairwise_nn_inner(self, x: torch.Tensor) -> torch.Tensor:
        # Pairwise dot products
        dots = torch.mm(x, x.t())
        n = x.size(0)
        dots.view(-1)[:: (n + 1)].fill_(-1)  # Trick to fill diagonal with -1

        # Max inner prod -> min distance
        ind = torch.argmax(dots, dim=1)

        return ind

    def forward(self, student_output: torch.Tensor, eps: float = 1e-8) -> torch.Tensor:
        with torch.amp.autocast(student_output.device.type, enabled=False):
            student_output = F.normalize(student_output, eps=eps, p=2, dim=-1)
            ind = self.pairwise_nn_inner(student_output)
            distances = self.pdist(student_output, student_output[ind])  # BxD, BxD -> B
            loss = -torch.log(distances + eps).mean()

        return loss


class DINOv2Student(SSLBaseNet):
    default_size = (224, 224)

    def __init__(
        self,
        backbone: PreTrainEncoder,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(backbone, config=config, size=size)
        assert self.config is not None, "must set config"
        assert isinstance(self.backbone, MaskedTokenRetentionMixin)

        dino_out_dim: int = self.config["dino_out_dim"]
        use_bn: bool = self.config["use_bn"]
        num_layers: int = self.config["num_layers"]
        hidden_dim: int = self.config["hidden_dim"]
        head_bottleneck_dim: int = self.config["head_bottleneck_dim"]
        ibot_separate_head: bool = self.config["ibot_separate_head"]
        ibot_out_dim: int = self.config.get("ibot_out_dim", dino_out_dim)

        self.dino_head = DINOHead(
            self.backbone.embedding_size,
            dino_out_dim,
            use_bn=use_bn,
            num_layers=num_layers,
            hidden_dim=hidden_dim,
            bottleneck_dim=head_bottleneck_dim,
        )
        if ibot_separate_head is False:
            self.ibot_head = None
        else:
            self.ibot_head = DINOHead(
                self.backbone.embedding_size,
                ibot_out_dim,
                use_bn=use_bn,
                num_layers=num_layers,
                hidden_dim=hidden_dim,
                bottleneck_dim=head_bottleneck_dim,
            )

        self.mask_token = nn.Parameter(torch.zeros(1, 1, 1, self.backbone.stem_width), requires_grad=True)

    # pylint: disable=arguments-differ
    def forward(  # type: ignore[override]
        self,
        global_crops: torch.Tensor,
        local_crops: torch.Tensor,
        mask: torch.Tensor,
        upper_bound: int,
        mask_indices_list: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        n_masked_patches = mask_indices_list.size(0)

        global_out = self.backbone.masked_encoding_retention(global_crops, mask, self.mask_token, return_keys="all")
        global_features = global_out["features"]
        global_features = global_features.flatten(2).transpose(1, 2)
        global_embedding = global_out["embedding"]

        local_embedding = self.backbone.embedding(local_crops)

        global_embedding_after_head = self.dino_head(global_embedding)
        local_embedding_after_head = self.dino_head(local_embedding)

        embed_dim = global_embedding.size(-1)
        buffer_tensor_patch_tokens = global_features.new_zeros(upper_bound, embed_dim)
        buffer_tensor_patch_tokens[:n_masked_patches].copy_(
            torch.index_select(global_features.flatten(0, 1), dim=0, index=mask_indices_list)
        )

        if self.ibot_head is None:
            global_masked_patch_tokens_after_head = self.dino_head(buffer_tensor_patch_tokens)[:n_masked_patches]
        else:
            global_masked_patch_tokens_after_head = self.ibot_head(buffer_tensor_patch_tokens)[:n_masked_patches]

        return (
            global_embedding,
            global_embedding_after_head,
            local_embedding_after_head,
            global_masked_patch_tokens_after_head,
        )


class DINOv2Teacher(SSLBaseNet):
    def __init__(
        self,
        backbone: PreTrainEncoder,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(backbone, config=config, size=size)
        assert self.config is not None, "must set config"
        assert isinstance(self.backbone, MaskedTokenRetentionMixin)

        dino_out_dim: int = self.config["dino_out_dim"]
        use_bn: bool = self.config["use_bn"]
        num_layers: int = self.config["num_layers"]
        hidden_dim: int = self.config["hidden_dim"]
        head_bottleneck_dim: int = self.config["head_bottleneck_dim"]
        ibot_separate_head: bool = self.config["ibot_separate_head"]
        ibot_out_dim: int = self.config.get("ibot_out_dim", dino_out_dim)

        self.dino_head = DINOHead(
            self.backbone.embedding_size,
            dino_out_dim,
            use_bn=use_bn,
            num_layers=num_layers,
            hidden_dim=hidden_dim,
            bottleneck_dim=head_bottleneck_dim,
        )
        if ibot_separate_head is False:
            self.ibot_head = None
        else:
            self.ibot_head = DINOHead(
                self.backbone.embedding_size,
                ibot_out_dim,
                use_bn=use_bn,
                num_layers=num_layers,
                hidden_dim=hidden_dim,
                bottleneck_dim=head_bottleneck_dim,
            )

        # Unused, Makes for an easier EMA update
        self.mask_token = nn.Parameter(torch.zeros(1, 1, 1, self.backbone.stem_width), requires_grad=True)

    # pylint: disable=arguments-differ
    def forward(  # type: ignore[override]
        self, x: torch.Tensor, n_crops: int, upper_bound: int, mask_indices_list: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        n_masked_patches = mask_indices_list.size(0)

        seq_len = (x.size(2) // self.backbone.max_stride) * (x.size(3) // self.backbone.max_stride)
        mask = torch.zeros([x.size(0), seq_len], device=x.device)

        out = self.backbone.masked_encoding_retention(x, mask=mask, return_keys="all")
        features = out["features"]
        features = features.flatten(2).transpose(1, 2)
        embedding = out["embedding"]

        embedding = embedding.chunk(n_crops)
        # NOTE: These are chunked and cat'd in reverse so A is matched to B in the global crops dino loss
        embedding = torch.concat((embedding[1], embedding[0]))

        patch_dim = features.size(-1)
        n = embedding.size(0)

        if self.ibot_head is None:
            buffer_tensor = features.new_zeros(upper_bound + n, patch_dim)
            buffer_tensor[:n].copy_(embedding)
            torch.index_select(
                features.flatten(0, 1),
                dim=0,
                index=mask_indices_list,
                out=buffer_tensor[n : n + n_masked_patches],
            )
            tokens_after_head = self.dino_head(buffer_tensor)
            embedding_after_head = tokens_after_head[:n]
            masked_patch_tokens_after_head = tokens_after_head[n : n + n_masked_patches]
        else:
            buffer_teacher = features.new_zeros(upper_bound, patch_dim)
            torch.index_select(
                features.flatten(0, 1),
                dim=0,
                index=mask_indices_list,
                out=buffer_teacher[:n_masked_patches],
            )
            embedding_after_head = self.dino_head(embedding)
            masked_patch_tokens_after_head = self.ibot_head(buffer_teacher)[:n_masked_patches]

        return (embedding_after_head, masked_patch_tokens_after_head)



import torch

from birder.kernels.load_kernel import load_soft_nms


class SoftNMS:
    """
    Soft-NMS -- Improving Object Detection With One Line of Code: https://arxiv.org/abs/1704.04503

    Lazy-loading Soft-NMS operator with batched processing support.

    The custom kernel is loaded on first instantiation, not at import time.
    Falls back to pure PyTorch implementation if kernel loading fails.

    Example:
        >>> soft_nms = SoftNMS()  # Kernel loads here
        >>> (keep, scores) = soft_nms(boxes, scores, idxs, sigma=0.5)
    """

    def __init__(self) -> None:
        if not torch.jit.is_tracing() and not torch.jit.is_scripting():
            self.soft_nms = load_soft_nms()
        else:
            self.soft_nms = None

    @property
    def is_available(self) -> bool:
        return self.soft_nms is not None

    def __call__(
        self,
        boxes: torch.Tensor,
        scores: torch.Tensor,
        idxs: torch.Tensor,
        sigma: float = 0.5,
        score_threshold: float = 0.1,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        if boxes.numel() == 0:
            return (
                torch.empty((0,), dtype=torch.int64, device=boxes.device),
                torch.empty((0,), dtype=torch.float32, device=scores.device),
            )

        # Offset boxes by category to prevent inter-category suppression
        max_coordinate = boxes.max()
        offsets = idxs.to(boxes) * (max_coordinate + 1)
        boxes_for_nms = boxes + offsets[:, None]

        if self.soft_nms is not None:
            return self.soft_nms.soft_nms(boxes_for_nms, scores, sigma, score_threshold)  # type: ignore[no-any-return]

        return _soft_nms(boxes_for_nms, scores, sigma, score_threshold)

    def soft_nms_single(
        self, boxes: torch.Tensor, scores: torch.Tensor, sigma: float = 0.5, score_threshold: float = 0.1
    ) -> tuple[torch.Tensor, torch.Tensor]:
        if self.soft_nms is not None:
            return self.soft_nms.soft_nms(boxes, scores, sigma, score_threshold)  # type: ignore[no-any-return]

        return _soft_nms(boxes, scores, sigma, score_threshold)


def _pairwise_iou(boxes1: torch.Tensor, boxes2: torch.Tensor) -> torch.Tensor:
    area1 = (boxes1[:, 2] - boxes1[:, 0]) * (boxes1[:, 3] - boxes1[:, 1])  # [N,]
    area2 = (boxes2[:, 2] - boxes2[:, 0]) * (boxes2[:, 3] - boxes2[:, 1])  # [M,]

    width_height = torch.min(boxes1[:, None, 2:], boxes2[:, 2:]) - torch.max(
        boxes1[:, None, :2], boxes2[:, :2]
    )  # [N,M,2]

    width_height.clamp_(min=0)  # [N,M,2]
    inter = width_height.prod(dim=2)  # [N,M]

    # handle empty boxes
    iou = torch.where(
        inter > 0,
        inter / (area1[:, None] + area2 - inter + 1e-8),
        torch.zeros(1, dtype=inter.dtype, device=inter.device),
    )

    return iou


def _soft_nms(
    boxes: torch.Tensor,
    scores: torch.Tensor,
    sigma: float = 0.5,
    score_threshold: float = 0.1,
) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Soft non-max suppression algorithm

    Taken from: https://github.com/rwightman/efficientdet-pytorch/blob/master/effdet/soft_nms.py
    Reference license: Apache-2.0
    """

    device = boxes.device
    boxes_remain = boxes.clone()
    scores_remain = scores.clone()
    num_elem = scores_remain.size()[0]
    idxs = torch.arange(num_elem, device=device)
    idxs_out = torch.zeros(num_elem, dtype=torch.int64, device=device)
    scores_out = torch.zeros(num_elem, dtype=torch.float32, device=device)
    count: int = 0

    while scores_remain.numel() > 0:
        top_idx = torch.argmax(scores_remain)
        idxs_out[count] = idxs[top_idx]
        scores_out[count] = scores_remain[top_idx]
        count += 1

        top_box = boxes_remain[top_idx]
        ious = _pairwise_iou(top_box.unsqueeze(0), boxes_remain)[0]

        decay = torch.exp(-torch.pow(ious, 2) / sigma)

        scores_remain *= decay
        keep = scores_remain > score_threshold
        keep[top_idx] = torch.tensor(False, device=device)

        boxes_remain = boxes_remain[keep]
        scores_remain = scores_remain[keep]
        idxs = idxs[keep]

    return (scores_out[:count], idxs_out[:count])


def batched_soft_nms(
    boxes: torch.Tensor,
    scores: torch.Tensor,
    idxs: torch.Tensor,
    sigma: float = 0.5,
    score_threshold: float = 0.1,
) -> tuple[torch.Tensor, torch.Tensor]:
    if boxes.numel() == 0:
        return (
            torch.empty((0,), dtype=torch.int64, device=boxes.device),
            torch.empty((0,), dtype=torch.float32, device=scores.device),
        )

    max_coordinate = boxes.max()
    offsets = idxs.to(boxes) * (max_coordinate + 1)
    boxes_for_nms = boxes + offsets[:, None]

    return _soft_nms(
        boxes_for_nms,
        scores,
        sigma=sigma,
        score_threshold=score_threshold,
    )
